###################################################################################################
# Setup Genomic Data Script
#   This script will fetch the necessary genomic data for all tools. Be aware that it will download
#   big data sets and it might be storage consuming. Also, this script will need an Internet
#   connection and may take several minutes to execute depending on the Internet speed.
###################################################################################################

# Import
from __future__ import print_function
import gzip
from optparse import OptionParser
from os import system, path, remove, mkdir
from sys import platform

# Optional Input
usage_message = "python setupGenomicData.py [options]"

# Initializing Option Parser
parser = OptionParser(usage=usage_message)

# Parameter: RGT Data Location
parser.add_option("--all", dest="all", action="store_true", default=False, help="Fetch all data sets.")
parser.add_option("--hg19", dest="hg19", action="store_true", default=False, help="Fetch human genome files.")
parser.add_option("--hg38", dest="hg38", action="store_true", default=False, help="Fetch human genome files.")
parser.add_option("--mm9", dest="mm9", action="store_true", default=False, help="Fetch mouse files.")
parser.add_option("--mm10", dest="mm10", action="store_true", default=False, help="Fetch mouse files.")
parser.add_option("--zv9", dest="zv9", action="store_true", default=False, help="Fetch zebrafish files.")
parser.add_option("--zv10", dest="zv10", action="store_true", default=False, help="Fetch zebrafish files.")

parser.add_option("--hg19-genome-path", type="string", metavar="STRING", dest="hg19_genome_path", default=None,
                  help="Path to an already existing hg19 genome (all chromosomes in the same file).")
parser.add_option("--hg38-genome-path", type="string", metavar="STRING", dest="hg38_genome_path", default=None,
                  help="Path to an already existing hg38 genome (all chromosomes in the same file).")
parser.add_option("--mm9-genome-path", type="string", metavar="STRING", dest="mm9_genome_path", default=None,
                  help="Path to an already existing mm9 genome (all chromosomes in the same file).")
parser.add_option("--mm10-genome-path", type="string", metavar="STRING", dest="mm10_genome_path", default=None,
                  help="Path to an already existing mm9 genome (all chromosomes in the same file).")
parser.add_option("--zv9-genome-path", type="string", metavar="STRING", dest="zv9_genome_path", default=None,
                  help="Path to an already existing zv9 genome (all chromosomes in the same file).")
parser.add_option("--zv10-genome-path", type="string", metavar="STRING", dest="zv10_genome_path", default=None,
                  help="Path to an already existing zv10 genome (all chromosomes in the same file).")

parser.add_option("--hg19-gtf-path", type="string", metavar="STRING", dest="hg19_gtf_path", default=None,
                  help="Path to an already existing hg19 GTF file.")
parser.add_option("--hg38-gtf-path", type="string", metavar="STRING", dest="hg38_gtf_path", default=None,
                  help="Path to an already existing hg38 GTF file.")
parser.add_option("--mm9-gtf-path", type="string", metavar="STRING", dest="mm9_gtf_path", default=None,
                  help="Path to an already existing mm9 GTF file.")
parser.add_option("--mm10-gtf-path", type="string", metavar="STRING", dest="mm10_gtf_path", default=None,
                  help="Path to an already existing mm10 GTF file.")
parser.add_option("--zv9-gtf-path", type="string", metavar="STRING", dest="zv9_gtf_path", default=None,
                  help="Path to an already existing zv9 GTF file.")
parser.add_option("--zv10-gtf-path", type="string", metavar="STRING", dest="zv10_gtf_path", default=None,
                  help="Path to an already existing zv10 GTF file.")

options, arguments = parser.parse_args()
if options.all:
    options.hg19 = True
    options.hg38 = True
    options.mm9 = True
    options.mm10 = True
    options.zv9 = True
    options.zv10 = True

###################################################################################################
# Parameters
###################################################################################################

# Current rgt data path
curr_dir = path.dirname(path.realpath(__file__))

# Platform
supported_platforms = ["linux", "linux2", "darwin"]
if platform not in supported_platforms:
    print("ERROR: This package currently supports only unix-based systems (Linux and MAC OS X).")
    exit(0)

###################################################################################################
# Genomic Data HG19
###################################################################################################

gencode_url = "ftp://ftp.sanger.ac.uk/pub/gencode/"

if options.hg19:

    output_location = path.join(curr_dir, "hg19")
    if not path.exists(output_location): mkdir(output_location)

    # Fetching genome
    output_genome_file_name = path.join(output_location, "genome_hg19.fa")
    if options.hg19_genome_path:
        print("Creating symbolic link to HG19 genome")
        system("ln -s " + options.hg19_genome_path + " " + output_genome_file_name)
        print("OK")
    else:
        gen_root_url = "http://hgdownload.cse.ucsc.edu/goldenPath/hg19/chromosomes/"
        chr_list = ["chr" + str(e) for e in range(1, 23) + ["X", "Y", "M"]]
        output_genome_file = open(output_genome_file_name, "w")
        for chr_name in chr_list:
            print("Downloading hg19 genome (" + chr_name + ")")
            gz_file_name = path.join(output_location, chr_name + ".fa.gz")
            if path.isfile(gz_file_name): remove(gz_file_name)
            system("wget " + gen_root_url + chr_name + ".fa.gz -P " + output_location)
            gz_file = gzip.open(gz_file_name, 'rb')
            output_genome_file.write(gz_file.read())
            gz_file.close()
            remove(gz_file_name)
            print("OK")
        output_genome_file.close()

    # Fetching GTF
    gtf_output_file_name = path.join(output_location, "gencode.v19.annotation.gtf")
    if options.hg19_gtf_path:
        print("Creating symbolic link to HG19 GTF")
        system("ln -s " + options.hg19_gtf_path + " " + gtf_output_file_name)
        print("OK")
    else:
        gtf_url = gencode_url + "Gencode_human/release_19/gencode.v19.annotation.gtf.gz"
        gtf_output_file_name_gz = path.join(output_location, "gencode.v19.annotation.gtf.gz")
        if path.isfile(gtf_output_file_name_gz): remove(gtf_output_file_name_gz)
        print("Downloading hg19 GTF (gene annotation) from genode")
        system("wget " + gtf_url + " -P " + output_location)
        gz_file = gzip.open(gtf_output_file_name_gz, 'rb')
        gtf_output_file = open(gtf_output_file_name, "w")
        gtf_output_file.write(gz_file.read())
        gz_file.close()
        remove(gtf_output_file_name_gz)
        gtf_output_file.close()
        print("OK")

###################################################################################################
# Genomic Data HG38
###################################################################################################

if options.hg38:

    output_location = path.join(curr_dir, "hg38")
    if not path.exists(output_location): mkdir(output_location)

    # Fetching genome
    output_genome_file_name = path.join(output_location, "genome_hg38.fa")
    if options.hg38_genome_path:
        print("Creating symbolic link to HG38 genome")
        system("ln -s " + options.hg38_genome_path + " " + output_genome_file_name)
        print("OK")
    else:
        gen_root_url = gencode_url + "Gencode_human/release_24/GRCh38.primary_assembly.genome.fa.gz"
        # chr_list = ["chr"+str(e) for e in range(1,23)+["X","Y","M"]]
        output_genome_file = open(output_genome_file_name, "w")
        # for chr_name in chr_list:
        print("Downloading hg38 genome")
        gz_file_name = path.join(output_location, "GRCh38.primary_assembly.genome.fa.gz")
        # if(path.isfile(gz_file_name)): remove(gz_file_name)
        system("wget " + gen_root_url + " -P " + output_location)
        gz_file = gzip.open(gz_file_name, 'rb')
        output_genome_file.write(gz_file.read())
        gz_file.close()
        remove(gz_file_name)
        print("OK")
        output_genome_file.close()

    # Fetching GTF
    gtf_output_file_name = path.join(output_location, "gencode.v24.annotation.gtf")
    if options.hg38_gtf_path:
        print("Creating symbolic link to HG38 GTF")
        system("ln -s " + options.hg38_gtf_path + " " + gtf_output_file_name)
        print("OK")
    else:
        gtf_url = gencode_url + "Gencode_human/release_24/gencode.v24.annotation.gtf.gz"
        gtf_output_file_name_gz = path.join(output_location, "gencode.v24.annotation.gtf.gz")
        if path.isfile(gtf_output_file_name_gz): remove(gtf_output_file_name_gz)
        print("Downloading hg19 GTF (gene annotation) from genode")
        system("wget " + gtf_url + " -P " + output_location)
        gz_file = gzip.open(gtf_output_file_name_gz, 'rb')
        gtf_output_file = open(gtf_output_file_name, "w")
        gtf_output_file.write(gz_file.read())
        gz_file.close()
        remove(gtf_output_file_name_gz)
        gtf_output_file.close()
        print("OK")

###################################################################################################
# Genomic Data MM9
###################################################################################################

if options.mm9:

    output_location = path.join(curr_dir, "mm9")
    if not path.exists(output_location): mkdir(output_location)

    # Fetching genome
    output_genome_file_name = path.join(output_location, "genome_mm9.fa")
    if options.mm9_genome_path:
        print("Creating symbolic link to MM9 genome")
        system("ln -s " + options.mm9_genome_path + " " + output_genome_file_name)
        print("OK")
    else:
        gen_root_url = "http://hgdownload.cse.ucsc.edu/goldenPath/mm9/chromosomes/"
        chr_list = ["chr" + str(e) for e in range(1, 20) + ["X", "Y", "M"]]
        output_genome_file = open(output_genome_file_name, "w")
        for chr_name in chr_list:
            print("Downloading MM9 genome (" + chr_name + ")")
            gz_file_name = path.join(output_location, chr_name + ".fa.gz")
            if path.isfile(gz_file_name): remove(gz_file_name)
            system("wget " + gen_root_url + chr_name + ".fa.gz -P " + output_location)
            gz_file = gzip.open(gz_file_name, 'rb')
            output_genome_file.write(gz_file.read())
            gz_file.close()
            remove(gz_file_name)
            print("OK")
        output_genome_file.close()

    # Fetching GTF
    gtf_output_file_name = path.join(output_location, "gencode.vM1.annotation.gtf")
    if options.mm9_gtf_path:
        print("Creating symbolic link to MM9 GTF")
        system("ln -s " + options.mm9_gtf_path + " " + gtf_output_file_name)
        print("OK")
    else:
        gtf_url = gencode_url + "Gencode_mouse/release_M1/gencode.vM1.annotation.gtf.gz"
        gtf_output_file_name_gz = path.join(output_location, "gencode.vM1.annotation.gtf.gz")
        if path.isfile(gtf_output_file_name_gz): remove(gtf_output_file_name_gz)
        print("Downloading MM9 GTF (gene annotation)")
        system("wget " + gtf_url + " -P " + output_location)
        gz_file = gzip.open(gtf_output_file_name_gz, 'rb')
        gtf_output_file = open(gtf_output_file_name, "w")
        gtf_output_file.write(gz_file.read())
        gz_file.close()
        remove(gtf_output_file_name_gz)
        gtf_output_file.close()
        print("OK")

###################################################################################################
# Genomic Data MM10
###################################################################################################

if options.mm10:

    output_location = path.join(curr_dir, "mm10")
    if not path.exists(output_location): mkdir(output_location)

    # Fetching genome
    output_genome_file_name = path.join(output_location, "genome_mm10.fa")
    if options.mm10_genome_path:
        print("Creating symbolic link to MM10 genome")
        system("ln -s " + options.mm10_genome_path + " " + output_genome_file_name)
        print("OK")
    else:
        gen_root_url = gencode_url + "Gencode_mouse/release_M11/GRCm38.primary_assembly.genome.fa.gz"
        output_genome_file = open(output_genome_file_name, "w")
        print("Downloading mm10 genome")
        gz_file_name = path.join(output_location, "GRCm38.primary_assembly.genome.fa.gz")
        system("wget " + gen_root_url + " -P " + output_location)
        gz_file = gzip.open(gz_file_name, 'rb')
        output_genome_file.write(gz_file.read())
        gz_file.close()
        remove(gz_file_name)
        print("OK")
        output_genome_file.close()

    # Fetching GTF
    gtf_output_file_name = path.join(output_location, "gencode.vM11.annotation.gtf")
    if options.mm10_gtf_path:
        print("Creating symbolic link to MM10 GTF")
        system("ln -s " + options.mm10_gtf_path + " " + gtf_output_file_name)
        print("OK")
    else:
        gtf_url = gencode_url + "Gencode_mouse/release_M11/gencode.vM11.annotation.gtf.gz"
        gtf_output_file_name_gz = path.join(output_location, "gencode.vM11.annotation.gtf.gz")
        if path.isfile(gtf_output_file_name_gz): remove(gtf_output_file_name_gz)
        print("Downloading MM10 GTF (gene annotation)")
        system("wget " + gtf_url + " -P " + output_location)
        gz_file = gzip.open(gtf_output_file_name_gz, 'rb')
        gtf_output_file = open(gtf_output_file_name, "w")
        gtf_output_file.write(gz_file.read())
        gz_file.close()
        remove(gtf_output_file_name_gz)
        gtf_output_file.close()
        print("OK")

###################################################################################################
# Genomic Data ZV9
###################################################################################################

if options.zv9:

    output_location = path.join(curr_dir, "zv9")
    if not path.exists(output_location): mkdir(output_location)

    # Fetching genome
    output_genome_file_name = path.join(output_location, "genome_zv9_ensembl_release_79.fa")
    if options.zv9_genome_path:
        print("Creating symbolic link to ZV9 genome")
        system("ln -s " + options.zv9_genome_path + " " + output_genome_file_name)
        print("OK")
    else:
        gen_root_url = "ftp://ftp.ensembl.org/pub/release-79/fasta/danio_rerio/dna/"
        chr_list = [str(e) for e in range(1, 25)] + ["MT"]
        output_genome_file = open(output_genome_file_name, "w")
        for chr_name in chr_list:
            print("Downloading ZV9 genome (chromosome " + chr_name + ")")
            gz_file_name = path.join(output_location, "Danio_rerio.Zv9.dna.chromosome." + chr_name + ".fa.gz")
            if path.isfile(gz_file_name): remove(gz_file_name)
            # Danio_rerio.Zv9.dna.chromosome.8.fa.gz
            system(
                "wget " + gen_root_url + "Danio_rerio.Zv9.dna.chromosome." + chr_name + ".fa.gz -P " + output_location)
            gz_file = gzip.open(gz_file_name, 'rb')
            output_genome_file.write(gz_file.read())
            gz_file.close()
            remove(gz_file_name)
            print("OK")
        output_genome_file.close()

    # Fetching GTF
    gtf_output_file_name = path.join(output_location, "Danio_rerio.Zv9.79.gtf")
    if options.zv9_gtf_path:
        print("Creating symbolic link to ZV9 GTF")
        system("ln -s " + options.zv9_gtf_path + " " + gtf_output_file_name)
        print("OK")
    else:
        # ftp://ftp.ensembl.org/pub/release-79/gtf/danio_rerio/Danio_rerio.Zv9.79.gtf.gz
        gtf_url = "ftp://ftp.ensembl.org/pub/release-79/gtf/danio_rerio/Danio_rerio.Zv9.79.gtf.gz"
        gtf_output_file_name_gz = path.join(output_location, "Danio_rerio.Zv9.79.gtf.gz")
        if path.isfile(gtf_output_file_name_gz): remove(gtf_output_file_name_gz)
        print("Downloading ZV9 GTF (gene annotation)")
        system("wget " + gtf_url + " -P " + output_location)
        gz_file = gzip.open(gtf_output_file_name_gz, 'rb')
        gtf_output_file = open(gtf_output_file_name, "w")
        gtf_output_file.write(gz_file.read())
        gz_file.close()
        remove(gtf_output_file_name_gz)
        gtf_output_file.close()
        # TODO add chr in each entry
        print("OK")

###################################################################################################
# Genomic Data ZV10
###################################################################################################

if options.zv10:

    output_location = path.join(curr_dir, "zv10")
    if not path.exists(output_location): mkdir(output_location)

    # Fetching genome
    output_genome_file_name = path.join(output_location, "genome_zv10_ensembl_release_84.fa")
    if options.zv10_genome_path:
        print("Creating symbolic link to ZV10 genome")
        system("ln -s " + options.zv10_genome_path + " " + output_genome_file_name)
        print("OK")
    else:
        gen_root_url = "ftp://ftp.ensembl.org/pub/release-84/fasta/danio_rerio/dna/"
        chr_list = [str(e) for e in range(1, 25)] + ["MT"]
        output_genome_file = open(output_genome_file_name, "w")
        for chr_name in chr_list:
            print("Downloading ZV10 genome (chromosome " + chr_name + ")")
            gz_file_name = path.join(output_location, "Danio_rerio.GRCz10.dna.chromosome." + chr_name + ".fa.gz")
            if path.isfile(gz_file_name): remove(gz_file_name)
            # Danio_rerio.GRCz10.dna.chromosome.8.fa.gz
            system(
                "wget " + gen_root_url + "Danio_rerio.GRCz10.dna.chromosome." + chr_name + ".fa.gz -P " + output_location)
            gz_file = gzip.open(gz_file_name, 'rb')
            output_genome_file.write(gz_file.read())
            gz_file.close()
            remove(gz_file_name)
            print("OK")
        output_genome_file.close()

    # Fetching GTF
    gtf_output_file_name = path.join(output_location, "Danio_rerio.GRCz10.84.gtf")
    if options.zv10_gtf_path:
        print("Creating symbolic link to ZV10 GTF")
        system("ln -s " + options.zv10_gtf_path + " " + gtf_output_file_name)
        print("OK")
    else:
        # ftp://ftp.ensembl.org/pub/release-84/gtf/danio_rerio/Danio_rerio.GRCz10.84.gtf.gz
        gtf_url = "ftp://ftp.ensembl.org/pub/release-84/gtf/danio_rerio/Danio_rerio.GRCz10.84.gtf.gz"
        gtf_output_file_name_gz = path.join(output_location, "Danio_rerio.GRCz10.84.gtf.gz")
        if path.isfile(gtf_output_file_name_gz): remove(gtf_output_file_name_gz)
        print("Downloading ZV10 GTF (gene annotation)")
        system("wget " + gtf_url + " -P " + output_location)
        gz_file = gzip.open(gtf_output_file_name_gz, 'rb')
        gtf_output_file = open(gtf_output_file_name, "w")
        gtf_output_file.write(gz_file.read())
        gz_file.close()
        remove(gtf_output_file_name_gz)
        gtf_output_file.close()
        # system("sed -i -e 's/^/chr/' "+gtf_output_file_name)
        # TODO add chr in each entry
        print("OK")



###################################################################################################
# Libraries
###################################################################################################

# Python
import warnings

warnings.filterwarnings("ignore")

# Internal
from ..Util import ErrorHandler


###################################################################################################
# Classes
###################################################################################################

class PileupRegion:
    """
    Represent an region in which a fragment pileup will be calculated.
    It is passed to pysam's functions (such as 'fetch') and then __call__
    gets called for every instance of the iterator that pysam's functions
    return.

    Authors: Eduardo G. Gusmao.

    Methods:

    __call__(alignment):
    It is called for every 'alignment' instance found by pysam's 'fetch' method.
    It loads self.vector with such alignment based on self.ext.
    """

    def __init__(self, start, end, downstream_ext, upstream_ext, forward_shift, reverse_shift):
        """ 
        Initializes PileupRegion.

        Variables:
        start -- Start point (in bp) of this PileupRegion (integer).
        end -- End point (in bp) of this PileupRegion (integer).
        length -- Length of this PileupRegion (integer).
        downstream_ext -- Number of bps to extend towards the downstream region (right for forward strand and left for reverse strand).
        upstream_ext -- Number of bps to extend towards the upstream region (left for forward strand and right for reverse strand).
        forward_shift -- Number of bps to shift the reads aligned to the forward strand.
        Can be a positive number for a shift towards the downstream region
        (towards the inside of the aligned read) and a negative number for a shift towards the upstream region.
        reverse_shift -- Number of bps to shift the reads aligned to the reverse strand.
        Can be a positive number for a shift towards the upstream region and a negative number for a shift towards the downstream region
        (towards the inside of the aligned read).
        vector -- Pileup vector. Each extended fragment will contribute with +1 to each
                  position of this vector, in which the length equals self.length (list).
        """
        self.start = start
        self.end = end
        self.length = end - start
        self.downstream_ext = downstream_ext
        self.upstream_ext = upstream_ext
        self.forward_shift = forward_shift
        self.reverse_shift = reverse_shift
        self.vector = [0.0] * self.length

    def __call__(self, alignment):
        try:
            if (not alignment.is_reverse):
                cut_site = alignment.pos + self.forward_shift
                if cut_site >= self.start and cut_site < self.end:
                    self.vector[cut_site - self.start] += 1.0
                #for i in range(max(alignment.pos + self.forward_shift - self.upstream_ext, self.start),
                #               min(alignment.pos + self.forward_shift + self.downstream_ext, self.end - 1)):
            else:
                cut_site = alignment.aend + self.reverse_shift - 1
                if cut_site >= self.start and cut_site < self.end:
                    self.vector[cut_site - self.start] += 1.0
                #for i in range(max(alignment.aend + self.reverse_shift - self.downstream_ext, self.start),
                #               min(alignment.aend + self.reverse_shift + self.upstream_ext, self.end - 1)):
                #    self.vector[i - self.start] += 1.0
        except Exception:
            pass

    """
    def __call__(self, alignment):
        ""
        Updates self.vector with the fragment obtained with the 'fetch' function from
        pysam package. This function will be called for all fragments fetched.

        Keyword arguments:
        alignment -- File location + name in HMM format. See manual for full description of such format.
        
        Return:
        None -- It updates self.vector only.
        ""

        try:
            # Forward strand update
            if not alignment.is_reverse:
                for i in range(max(alignment.pos,self.start),min(alignment.pos+self.ext,self.end-1)):
                    self.vector[i-self.start] += 1.0 
            # Reverse strand update
            else:
                for i in range(max(alignment.aend-self.ext,self.start),min(alignment.aend,self.end-1)):
                    self.vector[i-self.start] += 1.0 
        except Exception: pass

    def __call2__(self, alignment):
        try:
            if(not alignment.is_reverse):
                for i in range(max(alignment.pos-self.ext,self.start),min(alignment.pos+self.ext,self.end-1)):
                    self.vector[i-self.start] += 1.0 
            else:
                for i in range(max(alignment.aend-self.ext,self.start),min(alignment.aend+self.ext,self.end-1)):
                    self.vector[i-self.start] += 1.0
        except Exception: pass
    """



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
THOR detects differential peaks in multiple ChIP-seq profiles associated
with two distinct biological conditions.

Copyright (C) 2014-2016 Manuel Allhoff (allhoff@aices.rwth-aachen.de)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

@author: Manuel Allhoff
"""

from __future__ import print_function
from rgt.Util import npath


def get_data_block(filepath, feature):
    with open(filepath) as f:
        data = []
        read = False
        for line in f:
            line = line.strip()
            if line.startswith("#") and line == "#" + str(feature):
                read = True
            
            if line.startswith("#") and line != "#" + str(feature):
                read = False
            
            if not line.startswith("#") and read:
                data.append(line)
                
    if len(data) == 1 and not (feature == "rep1" or feature == "rep2" or feature == "inputs1" or feature == "inputs2"):
        return data[0]
    else:
        return data


def input_parser(filepath):
    bamfiles_1 = get_data_block(filepath, "rep1")
    bamfiles_1 = map(npath, bamfiles_1)

    bamfiles_2 = get_data_block(filepath, "rep2")
    bamfiles_2 = map(npath, bamfiles_2)

    # genome is optional, so if we get an empty list
    # we set it to None, otherwise we normalise the path
    genome = get_data_block(filepath, "genome")
    genome = npath(genome) if genome else None

    # the chrom sizes are not optional, but right now it's undefined
    # what happens if the user doesn't specify them, or specifies more
    # than one. So we just relay whatever we got from the file.
    chrom_sizes = npath(get_data_block(filepath, "chrom_sizes"))
    chrom_sizes = npath(chrom_sizes) if chrom_sizes else chrom_sizes

    inputs1 = get_data_block(filepath, "inputs1")
    inputs1 = map(npath, inputs1)

    inputs2 = get_data_block(filepath, "inputs2")
    inputs2 = map(npath, inputs2)

    dims = [len(bamfiles_1), len(bamfiles_2)]
    
    if not inputs1 and not inputs2:
        inputs = None
    else:
        inputs = inputs1 + inputs2

    return bamfiles_1 + bamfiles_2, genome, chrom_sizes, inputs, dims






#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
AnnotationSet
===================
AnnotationSet represent genomic annotation from genes.

"""

# Python
from __future__ import print_function

# Internal
from rgt.GenomicRegionSet import *
from rgt.Util import GenomeData, MotifData, AuxiliaryFunctions


class AnnotationSet:
    """This class represents genomic annotation from genes.

    *Keyword arguments:*

        - gene_source -- Gene source annotation. It will be used to create the gene_list element. It can be:
            - A matrix (list of lists): An AnnotationSet will be created based on such matrix.
            - A string representing a gtf file: An AnnotationSet will be created based on such gtf file.
            - A string representing an organism: An AnnotationSet will be created based on the gtf file for that organism in data.config file.

        - tf_source -- TF source annotation. After initialization, this object is mapped with gene_list. It can be:
            - A matrix (list of lists): Represents a final tf_list element.
            - A list of mtf files: The tf_list will be created based on all mtf files.
            - A list of repositories: The tf_list will be created based on the mtf files associated with such repositories in data.config.

        - alias_source -- Alias dictionary source annotation. It can be:
            - A dictionary: An alias dictionary will be created based on such dictionary.
            - A string representing a alias (txt) file: An alias dictionary will be created based on such txt file.
            - A string representing an organism: An alias dictionary will be created based on the txt file for that organism in data.config file.
    """

    class GeneField:
        """Gtf fields constants.

        *Constants:*

            - GENOMIC_REGION.
            - ANNOTATION_SOURCE.
            - FEATURE_TYPE.
            - GENOMIC_PHASE.
            - GENE_ID.
            - TRANSCRIPT_ID.
            - GENE_TYPE.
            - GENE_STATUS.
            - GENE_NAMES.
            - TRANSCRIPT_TYPE.
            - TRANSCRIPT_STATUS.
            - TRANSCRIPT_NAME.
            - LEVEL.
            - EXACT_GENE_MATCHES.
            - INEXACT_GENE_MATCHES.
        """

        # Gff Fields 
        GENOMIC_REGION = 0
        ANNOTATION_SOURCE = 1
        FEATURE_TYPE = 2
        GENOMIC_PHASE = 3

        # Gtf Fields
        GENE_ID = 4
        TRANSCRIPT_ID = 5
        GENE_TYPE = 6
        GENE_STATUS = 7
        GENE_NAMES = 8
        TRANSCRIPT_TYPE = 9
        TRANSCRIPT_STATUS = 10
        TRANSCRIPT_NAME = 11
        LEVEL = 12

        # Internal Fields
        EXACT_GENE_MATCHES = 13
        INEXACT_GENE_MATCHES = 14

    class TfField:
        """Mtf fields constants.

        *Constants:*

            - MATRIX_ID.
            - SOURCE.
            - VERSION.
            - GENE_NAMES.
            - GROUP.
            - EXACT_GENE_MATCHES.
            - INEXACT_GENE_MATCHES.
        """

        # Mtf Fields
        MATRIX_ID = 0
        SOURCE = 1
        VERSION = 2
        GENE_NAMES = 3
        GROUP = 4

        # Internal Fields
        EXACT_GENE_MATCHES = 5
        INEXACT_GENE_MATCHES = 6

    class DataType:
        """Data type constants.

        *Constants:*

            - GENE_LIST.
            - TF_LIST.
        """
        GENE_LIST = 0
        TF_LIST = 1

    class ReturnType:
        """Return type constants.

        *Constants:*

            - ANNOTATION_SET.
            - LIST.
        """
        ANNOTATION_SET = 0
        LIST = 1

    def __init__(self, gene_source, tf_source=None, alias_source=None, 
                 filter_havana=False, protein_coding=False, known_only=False):

        # Class Objects
        self.gene_list = [] # Represents gene annotation.
        self.tf_list = [] # Represents TF PWM annotation.
        self.alias_dict = dict() # Gene Symbol or other IDs -> ENSEMBL ID
        self.symbol_dict = dict() # ENSEMBL ID -> Official gene symbol

        # Initializing Required Field - Gene List
        if isinstance(gene_source,list): # It can be a matrix - Used by internal methods.
            self.gene_list = gene_source
        if isinstance(gene_source,str): # It can be a string.
            if os.path.isfile(gene_source): # The string may represent a path to a gtf file.
                # FTT for TDF True
                #filter_havana = False
                protein_coding = False
                known_only = False
                self.load_gene_list(gene_source, 
                                    filter_havana=filter_havana, 
                                    protein_coding=protein_coding,
                                    known_only=known_only)
            else: # The string may represent an organism which points to a gtf file within data.config.
                genome_data = GenomeData(gene_source)
                self.load_gene_list(genome_data.get_annotation(), 
                                    filter_havana=filter_havana, 
                                    protein_coding=protein_coding,
                                    known_only=known_only)

        # Initializing Optional Field - TF List
        if tf_source:
            if isinstance(tf_source, list):
                if isinstance(tf_source[0], list): # It can be a matrix
                    self.tf_list = tf_source
                else:
                    mtf_file_list = []
                    motif_data = MotifData()
                    for e in tf_source:
                        if os.path.isfile(e): # It can be a path to a mtf file.
                            mtf_file_list.append(e)
                        else: # It can represent an organism which points to an mtf file within data.config.
                            mtf_file = motif_data.get_mtf_path(e)
                            mtf_file_list.append(mtf_file)
                    self.tf_list = self.load_tf_list(mtf_file_list)
            else: pass # TODO Throw error.

        # Initializing Optional Field - Alias Dictionary
        if alias_source:
            if isinstance(alias_source,dict): # It can be a dictionary - Used by internal methods.
                self.alias_dict = alias_source
            if isinstance(alias_source,str): # It can be a string.
                if os.path.isfile(alias_source): # The string may represent a path to a txt alias file.
                    self.load_alias_dict(alias_source)
                else: # The string may represent an organism which points to a txt alias file within data.config.
                    genome_data = GenomeData(alias_source)
                    self.load_alias_dict(genome_data.get_gene_alias())
            else: pass # TODO Throw error

    def load_gene_list(self, file_name, filter_havana=True, protein_coding=False, known_only=False):
        """Reads gene annotation in gtf (gencode) format. It populates self.gene_list with such entries.
        
        *Keyword arguments:*

            - file_name -- The gencode .gtf file name.
        """
        # Opening GTF file
        try: gtf_file = open(file_name, "r")
        except Exception: 
            print("Error: Cannot find the annotation file: "+file_name)
            print("Please check the path in ~/rgtdata/data.config")
            sys.exit(1)
        
        # Reading GTF file
        for line in gtf_file:

            # Processing line
            line = line.strip()
            if line[0] == "#": continue
            line_list = line.split("\t")
            try:
                if filter_havana and line_list[1] == "HAVANA": continue
            except: pass
            
            addt_list = line_list[8].split(";")
            addt_list = filter(None, addt_list)

            # Processing additional list of options
            addt_dict = dict()
            for addt_element in addt_list:
                addt_element_list = addt_element.split(" ")
                addt_element_list = filter(None, addt_element_list)
                # Removing " symbol from string options
                addt_element_list[1] = addt_element_list[1].replace("\"", "")
                addt_dict[addt_element_list[0]] = addt_element_list[1]

            # filter non-protein-coding sequences, if required
            if protein_coding:
                if "gene_type" not in addt_dict or addt_dict["gene_type"] != "protein_coding":
                    continue
                if "transcript_type" in addt_dict and addt_dict["transcript_type"] != "protein_coding":
                    continue

            # filter unknown sequences, if required
            if known_only:
                if "gene_status" not in addt_dict or addt_dict["gene_status"] != "KNOWN":
                    continue
                if "transcript_status" in addt_dict and addt_dict["transcript_status"] != "KNOWN":
                    continue
        
            # Removing dot from IDs
            addt_dict["gene_id"] = addt_dict["gene_id"].split(".")[0]
            try: addt_dict["transcript_id"] = addt_dict["transcript_id"].split(".")[0]
            except: pass

            # Creating final version of additional arguments
            final_addt_list = []
            for addt_key in ["gene_id", "transcript_id", "gene_type", "gene_status", "gene_name", 
                             "transcript_type", "transcript_status", "transcript_name", "level"]:
                try:
                    final_addt_list.append(addt_dict[addt_key])
                except Exception:
                    final_addt_list.append(None)

            # Handling score
            current_score = 0
            if AuxiliaryFunctions.string_is_int(line_list[5]):
                current_score = AuxiliaryFunctions.correct_standard_bed_score(line_list[5])
            
            # Creating GenomicRegion
            genomic_region = GenomicRegion(chrom=line_list[0],
                                           initial=int(line_list[3])-1,
                                           final=int(line_list[4]),
                                           orientation=line_list[6],
                                           data=current_score)

            # Creating final vector
            extra_index_elements = [[],[]] # One list for each: EXACT_GENE_MATCHES, INEXACT_GENE_MATCHES
            final_vector = [genomic_region,line_list[1],line_list[2],line_list[7]] + final_addt_list + extra_index_elements
            self.gene_list.append(final_vector)

        # Termination
        gtf_file.close()

    def load_alias_dict(self, file_name):
        """Reads an alias.txt file and creates a dictionary to translate gene symbols/alternative IDs to ensembl gene ID
        
        *Keyword arguments:*

            - file_name -- Alias file name.
        """

        # Opening alias file
        alias_file = open(file_name,"r")

        # Iterating over alias file entries
        for line in alias_file:
            ll = line.strip().split("\t")
            ensembl_id = ll[0]
            official_name = ll[1]
            alias_vec = ll[2].split("&")
            self.symbol_dict[ensembl_id] = official_name
            for e in alias_vec:
                try: self.alias_dict[e].append(ensembl_id)
                except Exception: self.alias_dict[e] = [ensembl_id]

        # Termination
        alias_file.close()

    def load_tf_list(self, file_name_list):
        """Reads TF annotation in mtf (internal -- check manual) format. It populates self.tf_list with such entries. Everytime a TF annotation is loaded, a mapping with gene list is performed.
        
        *Keyword arguments:*

            - file_name_list -- A list with .mtf files.
        """

        # Iterating over the file name list
        for file_name in file_name_list:

            # Opening MTF file
            try: mtf_file = open(file_name,"r")
            except Exception: pass # TODO

            # Reading file
            for line in mtf_file:

                # Processing line
                line_list = line.strip().split("\t")
                while len(line_list) < 5: line_list.append(".")
                
                # Creating final vector
                extra_index_elements = [[],[]] # One list for each: EXACT_GENE_MATCHES, INEXACT_GENE_MATCHES
                final_vector = line_list + extra_index_elements
                self.tf_list.append(final_vector)

            # Termination
            mtf_file.close()

        # Mapping gene and tf lists
        self.map_lists()

    def map_lists(self):
        """Maps self.gene_list with self.tf_list in various ways."""
        self.exact_mapping()
        #self.inexact_mapping()

    def exact_mapping(self):
        """Maps (O(n log n)) exact entries of self.gene_list's gene names with self.tf_list's gene names."""
        
        # Sorting lists
        self.gene_list = sorted(self.gene_list, key=lambda k: k[self.GeneField.GENE_NAMES])
        tf_assist = []
        curr_tf_index = 0
        for e in self.tf_list:
            gene_name_list = ";".join(e[self.TfField.GENE_NAMES].split("+")).split(";")
            for k in gene_name_list:
                tf_assist.append([k,curr_tf_index])
            curr_tf_index += 1
        tf_assist = sorted(tf_assist, key=lambda k: k[0])

        # Initializing index
        curr_gene_index = 0
        curr_tf_index = 0

        # Linear while comparison loop
        while True:
            gene_genename = self.gene_list[curr_gene_index][self.GeneField.GENE_NAMES]
            tf_genename = tf_assist[0]
            gene_vec = self.gene_list[curr_gene_index]
            tf_vec = self.tf_list[tf_assist[1]]
            try: cmp_res = cmp(gene_genename,tf_genename)
            except IndexError: break
            if cmp_res == 0:
                gene_vec[self.GeneField.EXACT_GENE_MATCHES] = tf_vec
                tf_vec[self.TfField.EXACT_GENE_MATCHES] = gene_vec
                curr_gene_index += 1
            elif cmp_res == -1:
                curr_gene_index += 1
            elif cmp_res == 1:
                curr_tf_index += 1

    def inexact_mapping(self):
        """Comming soon!"""
        pass # TODO

    def fix_gene_names(self,gene_set,output_dict=False,mute_warn=False):
        """Checks if all gene names in gene_set are ensembl IDs. If a gene is not in ensembl format, it will be converted using alias_dict. If the gene name cannot be found then it is reported in a separate gene_set
        
        *Keyword arguments:*

            - gene_set -- A GeneSet object.
            - output_dict -- Also output the mapping dictionary (default = False).
            - mute_warn -- Do not print warnings regarding genes that mapped to multiple entries (default = False).
        
        *Return:*

            - mapped_gene_list -- A list of ensembl IDs
            - unmapped_gene_list -- A list of unmapped gene symbols/IDs
        """
        
        # Creating resulting lists
        mapped_gene_list = []
        unmapped_gene_list = []
        
        if output_dict: maping_dict = {}
        # Iterating on gene names
        for gene_name in gene_set.genes:
        
            # Verifying if it is ensembl name
            flag_ensembl = False
            try:
                if len(gene_name) >= 15 and gene_name[:3] == "ENS": flag_ensembl = True
            except Exception: flag_ensembl = False
            if flag_ensembl:
                mapped_gene_list.append(gene_name)
                sym = self.get_official_symbol(gene_name)
                if output_dict and sym and gene_name:
                    maping_dict[gene_name] = sym
            else:
                try:
                    alias_list = self.alias_dict[gene_name.upper()]
                    if len(alias_list) > 1:
                        if not mute_warn: 
                            print("Warning: The gene "+gene_name+" contains more than one matching IDs, both will be used.")
                    for e in alias_list: 
                        mapped_gene_list.append(e)
                        if output_dict and e and gene_name: 
                            maping_dict[e] = gene_name

                except Exception: unmapped_gene_list.append(gene_name)
        if output_dict:
            return mapped_gene_list, unmapped_gene_list, maping_dict
        else:
            return mapped_gene_list, unmapped_gene_list

    def get(self, query=None, list_type=DataType.GENE_LIST, return_type=ReturnType.ANNOTATION_SET):
        """Gets subsets of either self objects and returns different types.

        *Keyword arguments:*

            - query -- A parameter that allows for subsets of self to be fetched. It can be:
                - None: All fields/values are going to be returned.
                - A dictionary: Subsets the desired list according to this structure. Each
                      key must be a field (please refer to AnnotationSet.GeneField or AnnotationSet.TfField)
                      that must point to a single value or a list of values.
            - list_type -- Indicates which list should be subsetted/returned. Please refer to AnnotationSet.DataType.
            - return_type -- Indicates what should be returned. Please refer to AnnotationSet.ReturnType.

        *Return:*

            - result_list -- A <return_type> containing the requested <list_type> subsetted according to <query>.
        """

        # Fetching local copies of the lists
        gene_list = deepcopy(self.gene_list)
        tf_list = deepcopy(self.tf_list)
        current_list = None

        # Deciding which list to be subsetted/returned
        if list_type == self.DataType.GENE_LIST: current_list = gene_list
        elif list_type == self.DataType.TF_LIST: current_list = tf_list

        # Subsetting
        if isinstance(query, dict):
        
            # Iterating over query elements
            for query_key in query.keys():

                # O(n) operation if a single value is being queried.
                if not isinstance(query[query_key], list):
                    current_list = filter(lambda k: k[query_key] == query[query_key], current_list)

                # O(n log n) operation if multiple values are being queried.
                else:

                    # Sorting input lists
                    current_list = sorted(current_list, key=lambda k: k[query_key])
                    values_list = sorted(query[query_key])

                    # Initializing index and output structures
                    result_list = []
                    curr_value_index = 0
                    curr_list_index = 0

                    # Linear while comparison loop
                    while True:
                        try: cmp_res = cmp(current_list[curr_list_index][query_key], values_list[curr_value_index])
                        except IndexError: break
                        if cmp_res == 0:
                            result_list.append(current_list[curr_list_index])
                            curr_list_index += 1
                        elif cmp_res == -1:
                            curr_list_index += 1
                        elif cmp_res == 1:
                            curr_value_index += 1

                    # Resolving reference
                    current_list = result_list

        # Deciding which list to be subsetted/returned
        if list_type == self.DataType.GENE_LIST: gene_list = current_list
        elif list_type == self.DataType.TF_LIST: tf_list = current_list

        # Return
        if return_type == self.ReturnType.ANNOTATION_SET:
            # FIXME is deepcopy necessary here? seems overkill
            return AnnotationSet(gene_list, tf_source=tf_list, alias_source=deepcopy(self.alias_dict))
        elif return_type == self.ReturnType.LIST:
            return current_list

    def get_official_symbol(self, gene_name_source):
        """Returns the official symbol(s) from gene_name_source.

        *Keyword arguments:*

            - gene_source -- It can be a string (single gene name) or a GeneSet (multiple genes).

        *Return:*

            - if gene_source is string then returns the converted string gene name or None if gene name could not be converted.
            - if gene_source is list then returns two lists containing, respectively, converted and not-converted gene names.
        """

        if isinstance(gene_name_source, str):
            try: return self.symbol_dict[gene_name_source]
            except Exception: return None
        else:
            if isinstance(gene_name_source,list): curr_list = gene_name_source
            else: curr_list = gene_name_source.genes
            mapped_list = []
            unmapped_list = []
            for e in curr_list:
                try: mapped_list.append(self.symbol_dict[e])
                except:
                    try: mapped_list.append(self.symbol_dict[ e.split(".")[0] ])
                    except Exception: unmapped_list.append(e)
            return mapped_list, unmapped_list


    def get_promoters(self, promoterLength=1000, gene_set=None, unmaplist=False, variants=False):
        """
        Gets promoters of genes given a specific promoter length. It returns a GenomicRegionSet with such promoters.
        The ID of each gene will be put in the NAME field of each GenomicRegion.
        Each promoter includes also the coordinate of the 5' base pair, therefore each promoter actual
        length is promoterLength+1.

        *Keyword arguments:*

            - promoterLength -- The length of the promoter region.
            - gene_set -- A set of genes to narrow the search.
            - unmaplist -- If True than also return the unmappable genes list (default = False).

        *Return:*

            - result_grs -- A GenomicRegionSet containing the promoters.
            - unmapped_gene_list -- A list of genes that could not be mapped to an ENSEMBL ID.
        """

        # Fetching gene names
        mapped_gene_list = None
        unmapped_gene_list = None


        if gene_set: mapped_gene_list, unmapped_gene_list = self.fix_gene_names(gene_set)

        # Fetching genes

        if not variants: target = "gene"
        else: target = "transcript"
        if(gene_set): query_dictionary = {self.GeneField.FEATURE_TYPE:target, self.GeneField.GENE_ID:mapped_gene_list}
        else: query_dictionary = {self.GeneField.FEATURE_TYPE:target}
        
        query_annset = self.get(query_dictionary)

        # Creating GenomicRegionSet
        result_grs = GenomicRegionSet("promoters")
        for e in query_annset.gene_list:
            gr = e[self.GeneField.GENOMIC_REGION]
            if gr.orientation == "+":
                gr.final = gr.initial + 1
                gr.initial = gr.initial - promoterLength
            else:
                gr.initial = gr.final - 1
                gr.final = gr.initial + promoterLength + 1

            gr.name = e[self.GeneField.GENE_ID]
            result_grs.add(gr)
        if unmaplist: return result_grs, unmapped_gene_list
        else: return result_grs

    def get_tss(self, gene_set = None):
        """Gets TSS(Transcription start site) of genes. It returns a GenomicRegionSet with such TSS. The ID of each gene will be put in the NAME field of each GenomicRegion.

        *Keyword arguments:*

            - gene_set -- A set of genes to narrow the search.
        
        *Return:*

            - result_grs -- A GenomicRegionSet containing TSS.
            - unmapped_gene_list -- A list of genes that could not be mapped to an ENSEMBL ID.
        """

        # Fetching gene names
        mapped_gene_list = None
        unmapped_gene_list = None
        if gene_set: mapped_gene_list, unmapped_gene_list = self.fix_gene_names(gene_set)

        # Fetching genes
        if gene_set: query_dictionary = {self.GeneField.FEATURE_TYPE:"gene", self.GeneField.GENE_ID:mapped_gene_list}
        else: query_dictionary = {self.GeneField.FEATURE_TYPE:"gene"}
        query_annset = self.get(query_dictionary)

        # Creating GenomicRegionSet
        result_grs = GenomicRegionSet("TSS")
        for e in query_annset.gene_list:
            gr = e[self.GeneField.GENOMIC_REGION]
            if gr.orientation == "+":
                gr.initial = gr.initial
                gr.final = gr.initial + 1
            else:
                gr.initial = gr.final - 1
                gr.final = gr.final
            gr.name = e[self.GeneField.GENE_ID]
            result_grs.add(gr)
        result_grs.merge()
        if gene_set: return result_grs, unmapped_gene_list
        else: return result_grs

    def get_tts(self, gene_set = None):
        """Gets TTS(Transcription termination site) of genes. It returns a GenomicRegionSet with such TTS. The ID of each gene will be put in the NAME field of each GenomicRegion.

        *Keyword arguments:*

            - gene_set -- A set of genes to narrow the search.
        
        *Return:*

            - result_grs -- A GenomicRegionSet containing TTS.
            - unmapped_gene_list -- A list of genes that could not be mapped to an ENSEMBL ID.
        """

        # Fetching gene names
        mapped_gene_list = None
        unmapped_gene_list = None
        if gene_set: mapped_gene_list, unmapped_gene_list = self.fix_gene_names(gene_set)

        # Fetching genes
        if gene_set: query_dictionary = {self.GeneField.FEATURE_TYPE:"gene", self.GeneField.GENE_ID:mapped_gene_list}
        else: query_dictionary = {self.GeneField.FEATURE_TYPE:"gene"}
        query_annset = self.get(query_dictionary)

        # Creating GenomicRegionSet
        result_grs = GenomicRegionSet("TTS")
        for e in query_annset.gene_list:
            gr = e[self.GeneField.GENOMIC_REGION]
            if gr.orientation == "+":
                gr.initial = gr.initial
                gr.final = gr.initial + 1
            else:
                gr.initial = gr.final - 1
                gr.final = gr.final
            gr.name = e[self.GeneField.GENE_ID]
            result_grs.add(gr)
        result_grs.merge()
        if gene_set: return result_grs, unmapped_gene_list
        else: return result_grs
    
    def get_exons(self, start_site=False, end_site=False, gene_set = None, merge=True):
        """Gets exons of genes. It returns a GenomicRegionSet with such exons. The id of each gene will be put in the NAME field of each GenomicRegion.

        *Keyword arguments:*

            - start_site -- Whether to relocate the start sites.
            - end_site -- Whether to relocate the end sites.
            - gene_set -- A set of genes to narrow the search.

        *Return:*

            - result_grs -- A GenomicRegionSet containing the exons.
            - unmapped_gene_list -- A list of genes that could not be mapped to an ENSEMBL ID.
        """

        # Fetching gene names
        mapped_gene_list = None
        unmapped_gene_list = None
        if gene_set: mapped_gene_list, unmapped_gene_list = self.fix_gene_names(gene_set)

        # Fetching exons
        if gene_set: query_dictionary = {self.GeneField.FEATURE_TYPE:"exon", self.GeneField.GENE_ID:mapped_gene_list}
        else: query_dictionary = {self.GeneField.FEATURE_TYPE:"exon"}
        query_annset = self.get(query_dictionary)

        # Creating GenomicRegionSet
        result_grs = GenomicRegionSet("exon")
        for e in query_annset.gene_list:
            gr = e[self.GeneField.GENOMIC_REGION]
            #gr.name = e[self.GeneField.GENE_ID]
            gr.name = e[self.GeneField.TRANSCRIPT_ID]
            result_grs.add(gr)
        if start_site:
            result_grs.relocate_regions("leftend", left_length=1, right_length=1)
        elif end_site:
            result_grs.relocate_regions("rightend", left_length=1, right_length=1)
        if merge: result_grs.merge()
        if gene_set: return result_grs, unmapped_gene_list
        else: return result_grs
    
    def get_genes(self, gene_set = None):
        """Gets regions of genes. It returns a GenomicRegionSet with such genes. The id of each gene will be put in the NAME field of each GenomicRegion.

        *Keyword arguments:*

            - gene_set -- A set of genes to narrow the search.

        *Return:*

            - result_grs -- A GenomicRegionSet containing the genes.
            - unmapped_gene_list -- A list of genes that could not be mapped to an ENSEMBL ID.
        """

        # Fetching gene names
        mapped_gene_list = None
        unmapped_gene_list = None
        if gene_set: mapped_gene_list, unmapped_gene_list = self.fix_gene_names(gene_set)

        # Fetching genes
        if gene_set: query_dictionary = {self.GeneField.FEATURE_TYPE:"gene", self.GeneField.GENE_ID:mapped_gene_list}
        else: query_dictionary = {self.GeneField.FEATURE_TYPE:"gene"}
        query_annset = self.get(query_dictionary)

        # Creating GenomicRegionSet
        result_grs = GenomicRegionSet("genes")
        for e in query_annset.gene_list:
            gr = e[self.GeneField.GENOMIC_REGION]
            gr.name = e[self.GeneField.GENE_ID]
            result_grs.add(gr)
        result_grs.merge()
        if gene_set: return result_grs, unmapped_gene_list
        else: return result_grs
    
    def get_introns(self, start_site=False, end_site=False, gene_set = None):
        """Gets introns of genes. It returns a GenomicRegionSet with such introns. The id of each gene will be put in the NAME field of each GenomicRegion.

        *Keyword arguments:*

            - start_site -- Whether to relocate the start sites.
            - end_site -- Whether to relocate the end sites.
            - gene_set -- A set of genes to narrow the search.

        *Return:*

            - result_grs -- A GenomicRegionSet containing the introns.
            - unmapped_gene_list -- A list of genes that could not be mapped to an ENSEMBL ID.
        """

        if gene_set:
            genes, unmapped_gene_list = self.get_genes(gene_set=gene_set)
            exons, unmapped_gene_list_e = self.get_exons(gene_set=gene_set)
        else:
            genes = self.get_genes()
            exons = self.get_exons()
        
        result_grs = genes.subtract(exons)
        if start_site:
            result_grs.relocate_regions("leftend", left_length=1, right_length=1)
        elif end_site:
            result_grs.relocate_regions("rightend", left_length=1, right_length=1)
        if gene_set: return result_grs, unmapped_gene_list
        else: return result_grs

    def get_transcripts(self, gene_set = None):
        """Gets transcripts of genes. It returns a GenomicRegionSet with such transcripts. The id of each gene will be put in the NAME field of each GenomicRegion.

        *Keyword arguments:*

            - gene_set -- A set of genes to narrow the search.

        *Return:*

            - result_grs -- A GenomicRegionSet containing the exons.
            - unmapped_gene_list -- A list of genes that could not be mapped to an ENSEMBL ID.
        """

        # Fetching gene names
        mapped_gene_list = None
        unmapped_gene_list = None
        if gene_set: mapped_gene_list, unmapped_gene_list = self.fix_gene_names(gene_set)

        # Fetching exons
        if gene_set: query_dictionary = {self.GeneField.FEATURE_TYPE:"exon", self.GeneField.GENE_ID:mapped_gene_list}
        else: query_dictionary = {self.GeneField.FEATURE_TYPE:"exon"}
        query_annset = self.get(query_dictionary)

        # Creating GenomicRegionSet
        result_grs = GenomicRegionSet("exon")
        for e in query_annset.gene_list:
            gr = e[self.GeneField.GENOMIC_REGION]
            gr.name = e[self.GeneField.TRANSCRIPT_ID]
            result_grs.add(gr)

        if gene_set: return result_grs, unmapped_gene_list
        else: return result_grs



# Python Libraries
from __future__ import print_function
from collections import *
import numpy
import pysam
# Local Libraries

# Distal Libraries
from rgt.GenomicRegion import *
from rgt.GenomicRegionSet import *
from rgt.Util import OverlapType
from BindingSiteSet import BindingSite, BindingSiteSet

class RNADNABinding:
    """Describe a binding event of RNA and DNA to form a triplex 
    
    Attributes:
        self.rna        A BindingSite object which includes RNA name, \
                        region (start position and end position), \
                        score, error and its Sequence
        self.dna        A BindingSite object which includes DNA name, \
                        region (start position and end position), \
                        score, error and its Sequence
        self.score      The score from comparing RNA region and DNA region 
        self.err_rate   The ration of error base pair between RNA and DNA
        self.err        *************
        self.motif      The motif of triplex formation:
                        R - the purine motif that permit guanines (G) and adenines (A)
                        Y - the pyrimidine motif that permit cytosines (C) and thymines (T)
                        M - the mixed motif, purine-pyrimdine, that permit guanines (G) and thymines (T)
                        P - parallel binding, i.e. motifs facilitating Hoogsten bonds; 
                            this covers the pyrimidine motif and the purine-pyrimidine motif in parallel configuration
                        A - anti-parallel binding, i.e. motifs facilitating reverse Hoogsten bonds; 
                            this covers the purine motif and the purine-pyrimidine motif in anti-parallel configuration
        self.strand     The strand of DNA
        self.orient     The orientation of RNA regarding to DNA
        self.guan_rate  *************
        self.rna_seq    A string of RNA extended by gaps to match the position of DNA
        self.dna_seq    A string of DNA extended by gaps to match the position of RNA
        self.match      A string with '|' for perfect match, '*' for mismatch
    
    """

    __slots__ = ['rna', 'dna', 'motif', 'orient', 'score', 'err_rate', 'err', 'guan_rate', 
                 'rna_seq', 'dna_seq', 'match' ]

    def __init__(self, rna, dna, score, err_rate, err, guan_rate, rna_seq=None, dna_seq=None, match=None):
        """Initiation"""
        self.rna = rna  # BindingSite
        self.dna = dna  # GenomicRegion
        self.motif = rna.motif
        self.strand = dna.orientation
        self.orient = rna.orientation
        self.score = score
        self.err_rate = err_rate
        self.err = err
        self.guan_rate = guan_rate
        self.rna_seq = rna_seq
        self.dna_seq = dna_seq
        self.match = match
    
    def __str__(self):
        return "\t".join( [ self.rna.chrom, str(self.rna.initial), str(self.rna.final), self.dna.toString(), "0", 
                            str(len(self.dna)), self.score, self.err_rate, self.err, self.motif,
                            self.strand, self.orient, self.guan_rate ] )

    def __eq__(self, other):
        return (self.dna, self.rna) == (other.dna, other.rna)
        
class RNADNABindingSet:
    """A framework to map DNA binding sites to corresponding RNA binding sites 
    
    Attributes:
        self.rna    A BindingSequence-ID   
        [1] TFO start   
        [2] TFO end 
        [3] Duplex-ID   
        [4] TTS start   
        [5] TTS end 
        [6] Score   
        [7] Error-rate  
        [8] Errors  
        [9] Motif   
        [10] Strand  
        [11] Orientation 
        [12] Guanine-rate
    
    """
    __slots__ = ['name', 'sequences', 'sorted_dna', 'sorted_rna' ]

    def __init__(self, name):
        """Initialize"""
        self.name = name       # RNA name
        self.sequences = []    # A list to contain all RNA/DNA interactions and extra information 
        self.sorted_dna = False
        self.sorted_rna = False
 

    def __len__(self):
        """Return the number of interactions between DNA and RNA """
        return len(self.sequences)

    def __iter__(self):
        return iter(self.sequences)
    
    def add(self, rnadnabinding):
        """Add one pair of BindingSite on RNA and GenomicRegion on DNA"""
        self.sequences.append(rnadnabinding)
        
    def get_rbs(self, sort=False, orientation=None):
        """Return a BindingSiteSet which contains all RNA binding sites"""
        rna_set = BindingSiteSet(name=self.name)
        for rd in self.sequences:
            if not orientation:
                rna_set.add(rd.rna)
            elif orientation == rd.orient:
                rna_set.add(rd.rna)
            else: pass

        if sort: rna_set.sort()
        return rna_set

    def get_dbs(self, sort=False, orientation=None, rm_duplicate=False, dbd_tag=False):
        """Return GenomicRegionSet which contains all DNA binding sites"""
        dna_set = GenomicRegionSet(name="DNA_binding_sites")
        if len(self) == 0: return dna_set
        for rd in self.sequences:
            if dbd_tag:
                dbs = GenomicRegion(chrom=rd.dna.chrom, initial=rd.dna.initial, final=rd.dna.final,
                                    name=rd.rna.str_rna(), orientation=rd.dna.orientation, 
                                    data=rd.score)
            else:
                dbs = GenomicRegion(chrom=rd.dna.chrom, initial=rd.dna.initial, final=rd.dna.final,
                                    name=rd.dna.name, orientation=rd.dna.orientation, 
                                    data=rd.score)

            if not orientation:
                dna_set.add(dbs)
            else:
                if orientation == rd.orient:
                    dna_set.add(dbs)
                else: pass
        if sort: dna_set.sort()
        if rm_duplicate: dna_set.remove_duplicates()
        return dna_set

    def sort_rbs(self):
        """Sort the dictionary by RNA"""
        self.sequences = sorted(self.sequences, key=lambda x: x.rna, cmp=GenomicRegion.__cmp__)
        self.sorted_rna = True
        
    def sort_dbs(self):
        """Sort the dictionary by DNA"""
        self.sequences = sorted(self.sequences, key=lambda x: x.dna, cmp=GenomicRegion.__cmp__)
        self.sorted_dna = True
    
    def sort_dbs_by_regions(self, regionset):
        """Sort the DBS by given GenomicRegionSet"""
        dbss = self.get_dbs(sort=True)

        result = {}

        if not regionset.sorted: regionset.sort()
        
        iter_dbs = iter(dbss)
        dbs = iter_dbs.next()

        last_j = len(regionset)-1
        j = 0
        cont_loop = True
        pre_inter = 0
        cont_overlap = False

        while cont_loop:
            # When the regions overlap
            if dbs.overlap(regionset[j]):
                result[regionset[j].toString()].add(dbs)

                if cont_overlap == False: pre_inter = j
                if j == last_j: 
                    try: dbs = iter_dbs.next()
                    except: cont_loop = False 
                else: 
                    j = j + 1
                    result[regionset[j].toString()] = GenomicRegionSet("RBS_"+regionset[j].toString())
                cont_overlap = True
            
            elif dbs < regionset[j]:
                try: 
                    dbs = iter_dbs.next()
                    j = pre_inter
                    cont_overlap = False
                except: cont_loop = False 
            
            elif dbs > regionset[j]:
                if j == last_j:
                    cont_loop = False
                else:
                    j = j + 1
                    result[regionset[j].toString()] = GenomicRegionSet("RBS_"+regionset[j].toString())
                    cont_overlap = False
        return result

    def sort_rbs_by_regions(self, regionset, merge_rbs=True):
        """Sort the DBS by given GenomicRegionSet"""
        
        result = OrderedDict()

        #txp_copy = copy.deepcopy(self)
        #txp_copy.sort_dbs()
        self.sort_dbs()

        if not regionset.sorted: regionset.sort()

        for region in regionset:
            result[region.toString()] = BindingSiteSet("binding site:"+region.toString())
        
        if len(self) == 0: 
            return result
        #iter_rd = iter(txp_copy)
        iter_rd = iter(self)
        rd = iter_rd.next()

        last_j = len(regionset)-1
        j = 0
        cont_loop = True
        pre_inter = 0
        cont_overlap = False

        while cont_loop:
            # When the regions overlap
            if rd.dna.overlap(regionset[j]):
                result[regionset[j].toString()].add(rd.rna)

                if cont_overlap == False: pre_inter = j
                if j == last_j: 
                    try: rd = iter_rd.next()
                    except: cont_loop = False 
                else: 
                    j = j + 1
                    
                cont_overlap = True
            
            elif rd.dna < regionset[j]:
                try: 
                    rd = iter_rd.next()
                    j = pre_inter
                    cont_overlap = False
                except: cont_loop = False 
            
            elif rd.dna > regionset[j]:
                if j == last_j:
                    cont_loop = False
                else:
                    j = j + 1
                    cont_overlap = False
         

        return result


    def sort_rd_by_regions(self, regionset):
        """Sort RNADNA binding information by a given GenomicRegionSet"""
        """Sort the DBS by given GenomicRegionSet"""

        result = OrderedDict()

        #txp_copy = copy.deepcopy(self)
        #txp_copy.sort_dbs()
        self.sort_dbs()

        if not regionset.sorted: regionset.sort()

        for region in regionset:
            result[region.toString()] = RNADNABindingSet("RNADNA_interaction:"+region.toString())

        if len(self) == 0: 
            return result
        #iter_rd = iter(txp_copy)
        iter_rd = iter(self)
        
        rd = iter_rd.next()

        last_j = len(regionset)-1
        j = 0
        cont_loop = True
        pre_inter = 0
        cont_overlap = False
        

        while cont_loop:
            # When the regions overlap
            if rd.dna.overlap(regionset[j]):
                result[regionset[j].toString()].add(rd)

                if cont_overlap == False: pre_inter = j
                if j == last_j: 
                    try: rd = iter_rd.next()
                    except: cont_loop = False 
                else: 
                    j = j + 1
                cont_overlap = True
            
            elif rd.dna < regionset[j]:
                try: 
                    rd = iter_rd.next()
                    j = pre_inter
                    cont_overlap = False
                except: cont_loop = False 
            
            elif rd.dna > regionset[j]:
                if j == last_j:
                    cont_loop = False
                else:
                    j = j + 1
                    cont_overlap = False
        return result

    def sort_rd_by_rbs(self, rbss):
        """Sort RNADNA binding information by a given BindingSiteSet"""
        
        result = OrderedDict()

        #txp_copy = copy.deepcopy(self)
        #txp_copy.sort_rbs()

        
        self.sort_rbs()

        if not rbss.sorted: rbss.sort()

        #iter_rd = iter(txp_copy)
        iter_rd = iter(self)
        rd = iter_rd.next()

        last_j = len(rbss)-1
        j = 0
        cont_loop = True
        pre_inter = 0
        cont_overlap = False

        while cont_loop:
            # When the regions overlap
            if rd.rna.overlap(rbss[j]):
                result[rbss[j].str_rna()].add(rd)

                if cont_overlap == False: pre_inter = j
                if j == last_j: 
                    try: rd = iter_rd.next()
                    except: cont_loop = False 
                else: 
                    j = j + 1
                    result[rbss[j].str_rna()] = RNADNABindingSet("RNADNA_interaction:"+rbss[j].str_rna())
                cont_overlap = True
            
            elif rd.dna < rbss[j]:
                try: 
                    rd = iter_rd.next()
                    j = pre_inter
                    cont_overlap = False
                except: cont_loop = False 
            
            elif rd.dna > rbss[j]:
                if j == last_j:
                    cont_loop = False
                else:
                    j = j + 1
                    result[rbss[j].str_rna()] = RNADNABindingSet("RNADNA_interaction:"+rbss[j].str_rna())
                    cont_overlap = False
        return result


    def merge_rbs(self, rbss=None, rm_duplicate=False, asgene_organism=None, region_set=None, 
                  name_replace=None, cutoff=0):
        """Merge the RNA binding regions which have overlap to each other and 
           combine their corresponding DNA binding regions.
        
        extend -> Define the extending length in basepair of each RNA binding regions
        perfect_match -> Merge only the exactly same RNA binding regions
        """
        # A dict: RBS as key, and GenomicRegionSet as its value
        self.merged_dict = OrderedDict()
        # reg = copy.deepcopy(region_set)
        # res = copy.deepcopy(name_replace)
        # reg = region_set
        res = name_replace

        if not rbss:
            # Merge RBS
            rna_merged = self.get_rbs()
            rna_merged.merge()
        else:
            rna_merged = rbss

        for r in rna_merged:
            self.merged_dict[r] = GenomicRegionSet(r.toString())

        rbsm = iter(rna_merged)
        try: r = rbsm.next()
        except: return
                
        if self.sorted_rna: pass
        else: self.sort_rbs()

        con = iter(self)
        try: rd = con.next()
        except: return

        con_loop = True
        while con_loop:
            #print(".", end="")
            if r.overlap(rd.rna):
                self.merged_dict[r].add(rd.dna)
                try: rd = con.next()
                except: 
                    try:
                        r = rbsm.next()
                        # self.merged_dict[r] = GenomicRegionSet(r.toString())
                    except:
                        if rm_duplicate: self.merged_dict[r].remove_duplicates()
                        con_loop = False
            elif rd.rna < r:
                try: rd = con.next()
                except: 
                    try:
                        r = rbsm.next()
                        # self.merged_dict[r] = GenomicRegionSet(r.toString())
                    except: 
                        if rm_duplicate: self.merged_dict[r].remove_duplicates()
                        con_loop = False
            elif rd.rna > r:
                if rm_duplicate: self.merged_dict[r].remove_duplicates()
                    #print(self.merged_dict[r].sequences[0].name)
                try:
                    r = rbsm.next()
                    # self.merged_dict[r] = GenomicRegionSet(r.toString())
                except: 
                    try: rd = con.next()
                    except: con_loop = False

        if region_set:
            for r in self.merged_dict.keys():
                s = region_set.intersect(self.merged_dict[r],
                                         mode=OverlapType.ORIGINAL,
                                         rm_duplicates=rm_duplicate)
                self.merged_dict[r] = s

        if not region_set and rm_duplicate:
            for r in self.merged_dict.keys():
                self.merged_dict[r].remove_duplicates()

        if cutoff:
            for r in self.merged_dict.keys():
                if len(self.merged_dict[r]) < cutoff:
                    n = self.merged_dict.pop(r, None)
        if res:
            for r in self.merged_dict.keys():
                self.merged_dict[r].replace_region_name(regions=res)
        #self.merged_dict = new_dict
        elif asgene_organism:
            for r in self.merged_dict.keys():
                try: self.merged_dict[r] = self.merged_dict[r].gene_association(organism=asgene_organism)
                except: pass

    def read_txp(self, filename, dna_fine_posi=False, shift=None):
        """Read txp file to load all interactions. """
        
        with open(filename) as f:
            for line in f:
                # skip the comment line
                if line.startswith("TFO:") or line.startswith("TTS:") or line.startswith("    ") or line.startswith("#"):
                    continue
                if "\tchrM:" in line: continue # skip chromosome Mitocondria
                line = line.strip("\n")
                line = line.split("\t")
                
                if len(line) < 10: 
                    #print(line)
                    continue # skip the unimportant lines in txp

                if len(line) == 12:
                    line.insert(8,"_")
                # Format of each line in txp
                #     [0] Sequence-ID   
                #     [1] TFO start   
                #     [2] TFO end 
                #     [3] Duplex-ID   
                #     [4] TTS start   
                #     [5] TTS end 
                #     [6] Score   
                #     [7] Error-rate  
                #     [8] Errors  
                #     [9] Motif   
                #     [10] Strand  
                #     [11] Orientation 
                #     [12] Guanine-rate

                # RNA binding site
                if not self.name: self.name = line[0]
                
                if shift:
                    rna_start, rna_end = int(line[1])+shift, int(line[2])+shift
                else:
                    rna_start, rna_end = int(line[1]), int(line[2])
                
                if rna_start > rna_end: rna_start, rna_end =  rna_end, rna_start

                rna = BindingSite(chrom=line[0], initial=rna_start, final=rna_end, score=line[6], 
                                  errors_bp=line[8], motif=line[9], orientation=line[11], 
                                  guanine_rate=line[12])
                # DNA binding site
                rg = line[3].split(":")[1].split("-")
                try: rg.remove("")
                except: pass
                
                if dna_fine_posi:
                    dna_start = int(rg[0]) + int(line[4])
                    dna_end = int(rg[0]) + int(line[5])
                    dna = GenomicRegion(chrom=line[3].split(":")[0], initial=dna_start, final=dna_end, 
                                        name=line[3], orientation=line[10],
                                        data=[line[6], line[9], line[11]]) # score, motif, orientation

                else:
                    try:
                        dna = GenomicRegion(chrom=line[3].split(":")[0], initial=int(rg[0]), final=int(rg[1]), 
                                            name=line[3], orientation=line[10],
                                            data=[line[6], line[9], line[11]])
                    except:
                        print(line)
                # Map RNA binding site to DNA binding site
                self.add(RNADNABinding(rna=rna, dna=dna, score=line[6], err_rate=line[7], err=line[8], guan_rate=line[12]))

    def map_promoter_name(self, promoters):
        """Give each DNA region the corresponding name from the given promoter"""
        self.sort_dbs()
        if not promoters.sorted: promoters.sort()
        con_p = iter(promoters)
        p = con_p.next()

        for rd in self:
            if p.overlap(rd.dna):
                rd.dna.name = p.name
            elif p < rd.dna:
                try: p = con_p.next()
                except: break

    def write_txp(self, filename):
        """Write RNADNABindingSet into the file"""
        d = os.path.dirname(filename)
        if d:
            try: os.stat(d)
            except: os.mkdir(d)
        with open(filename, "w") as f:
            print("# RNA-ID\tRBS-start\tRBS-end\tDNA-ID\tDBS-start\tDBS-end\tScore\tError-rate\tErrors\
                   \tMotif\tStrand\tOrientation\tGuanine-rate", file=f)
            for rd in self:
                print(str(rd), file=f)

    def write_bed(self, filename, dbd_tag=True, remove_duplicates=False, convert_dict=None, associated=False):
        """Write BED file for all the DNA Binding sites
        filename: define the output filename
        remove_duplicates: remove all exact duplicates
        convert_dict: given a dictionary to change the region name
        """
        dbss = self.get_dbs(dbd_tag=dbd_tag)
        if remove_duplicates:
            dbss.remove_duplicates()
        if convert_dict:
            dbss = dbss.change_name_by_dict(convert_dict=convert_dict)
        if associated:
            dbss.add_associated_gene_data(organism=associated)
        dbss.write_bed(filename)


    def count_dbs_on_dbd(self, rna_len):
        """Return the frequency list of the appearance of DBS on RNA profile"""
        count = [0]*rna_len

        for rd in self.sequences:
            for i in range(rd.rna.initial, rd.dna.final):
                count[i] += 1

        return count

    def remove_duplicates(self):
        """Remove the duplicate RNA-DNA interactions and remain the unique ones. (No return)"""
        if self.sorted_dna == False: self.sort_dbs()
        
        for i in range(len(self.sequences) - 1):
            loop = True
            while loop:
                try:
                    if self.sequences[i] == self.sequences[i+1]:
                        del self.sequences[i+1]
                        loop = True
                    else:
                        loop = False
                except:
                    loop = False
                    continue

    def remove_duplicates_by_dbs(self):
        """Remove the duplicate RNA-DNA interactions and remain the unique ones. (No return)"""
        if self.sorted_dna == False: self.sort_dbs()
        
        for i in range(len(self.sequences) - 1):
            loop = True
            while loop:
                try:
                    if self.sequences[i].dna.toString() == self.sequences[i+1].dna.toString():
                        del self.sequences[i+1]
                        loop = True
                    else:
                        loop = False
                except:
                    loop = False
                    continue

    def generate_jaspar_matrix(self, genome_path, rbss=None):
        
        nucl = {"A":0 , "C":1 , "G":2 , "T":3 ,
                "a":0 , "c":1 , "g":2 , "t":3 }
        genome = pysam.Fastafile(genome_path)
        # A dict: RBS as key, and matrix as its value
        self.pwm_dict = OrderedDict()
        #print("merge_rbs")

        if not rbss:
            # Merge RBS
            rna_merged = self.get_rbs()
            rna_merged.merge()
        else:
            rna_merged = rbss

        for r in rna_merged:
            # Two same size matrix: P & A
            self.pwm_dict[r] = [ numpy.zeros((4, len(r))), numpy.zeros((4, len(r))) ]

        rbsm = iter(rna_merged)
        try: r = rbsm.next()
        except: return
                
        if self.sorted_rna: pass
        else: self.sort_rbs()

        con = iter(self)
        try: rd = con.next()
        except: return

        con_loop = True
        while con_loop:
            #print(".", end="")
            if r.overlap(rd.rna):
                seq = genome.fetch(rd.dna.chrom, max(0, rd.dna.initial), rd.dna.final)
                if rd.orient == "A": 
                    pa = 1
                    seq = seq[::-1]
                else: 
                    pa = 0
                print(seq)
                ind = rd.rna.initial - r.initial
                for i, a in enumerate(seq):
                    self.pwm_dict[r][pa][ nucl[a], ind + i ] += 1

                try: rd = con.next()
                except: 
                    try:
                        r = rbsm.next()
                    except:
                        con_loop = False
            elif rd.rna < r:
                try: rd = con.next()
                except: 
                    try:
                        r = rbsm.next()
                    except: 
                        con_loop = False
            elif rd.rna > r:
                try:
                    r = rbsm.next()
                except: 
                    try: rd = con.next()
                    except: con_loop = False

    def overlap_rbss(self, rbss):
        z = RNADNABindingSet(self.name)
        for rd in self.sequences:
            for rbs in rbss:
                if rd.rna.overlap(rbs):
                    z.add(rd)
        return z



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Copyright (C) 2014-2016 Manuel Allhoff (allhoff@aices.rwth-aachen.de)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

Estimate extension size of reads for an ChIP-seq experiment.

Author: Manuel Allhoff

Calculate the cross correlation between coverage of forward and backward reads
on first chromosome.
Methods is based on:

Kharchenko et al., Design and analysis of ChIP-seq experiments for DNA-binding
proteins (2008)

Methods:

get_extension_size(bamfile)
Return shift/extension size of reads descriebed by BAM file.

@author: Manuel Allhoff

"""

from __future__ import print_function
import pysam

cov_f = {}
cov_r = {}


def get_value(d, x):
    """Return d[x] of dictionary d or 0"""
    return d[x] if d.has_key(x) else 0


def get_read_size(filename):
    f = pysam.Samfile(filename, "rb")

    s = []
    i = 0
    for read in f.fetch():
        i += 1
        if i == 1000:
            break
        if not read.is_unmapped:
            s.append(read.rlen)

    # print('read length is %s' %str(sum(s)/len(s)))

    return sum(s) / len(s)


def init_cov(filename):
    file = pysam.Samfile(filename, "rb")

    for read in file.fetch(file.references[0]):
        if not read.is_unmapped:
            if not read.seq:
                h = 0
            else:
                h = len(read.seq)
            pos = read.pos + read.rlen - h if read.is_reverse else read.pos
            if read.is_reverse:
                if not cov_r.has_key(pos):
                    cov_r[pos] = 1
            else:
                if not cov_f.has_key(pos):
                    cov_f[pos] = 1


def ccf(k):
    """Return value of cross-correlation function"""
    s = 0
    forward_keys = set(cov_f.keys())
    reverse_keys = set(map(lambda x: x - k, cov_r.keys()))
    keys = forward_keys & reverse_keys

    for p in keys:
        s += get_value(cov_f, p) & get_value(cov_r, p + k)

    return s, k


def get_extension_size(filename, start=0, end=600, stepsize=5):
    """Return extension/shift size of reads and all computed values of the convolution. 
    Search value with a resolution of <stepsize> from start to end."""
    read_length = get_read_size(filename)
    start -= read_length

    init_cov(filename)

    r = map(ccf, range(start, end, stepsize))

    r = map(lambda x: (x[0], x[1]), r)

    # print('extension size is %s' %max(r[read_length/stepsize*2:])[1])

    return max(r[read_length / stepsize * 2:])[1], r


if __name__ == '__main__':
    # a, b = get_extension_size('/home/manuel/workspace/cluster_p/blueprint/raw/input/C000S5H1.Input.bwa_filtered.20130415.bam')
    a, b = get_extension_size('/home/manuel/workspace/cluster_p/allhoff/project_THOR/data/payton/FL1_H3K27ac_input.bam')
    print(a, b)

    for el in b:
        print(el[1], el[0], sep='\t')





import os
import sys
import io
import re
from shutil import copy
from pwd import getpwnam
from sys import platform, exit
from distutils import dir_util
from setuptools import setup, find_packages
from os import walk, chown, chmod, path, getenv, makedirs, remove
from optparse import OptionParser, BadOptionError, AmbiguousOptionError

if not sys.version_info[0] == 2:
    sys.exit("Sorry, Python 3 is not supported (yet)")

"""
Installs the RGT tool with standard setuptools options and additional
options specific for RGT.

Authors: Manuel Allhoff, Ivan G. Costa, Eduardo G. Gusmao, Joseph C.C. Kuo, Fabio Ticconi.

Installs the RGT tool with standard setuptools options and additional
options specific for RGT.
"""

def read(*names, **kwargs):
    with io.open(
        os.path.join(os.path.dirname(__file__), *names),
        encoding=kwargs.get("encoding", "utf8")
    ) as fp:
        return fp.read()


def find_version(*file_paths):
    version_file = read(*file_paths)
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]",
                              version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")

current_version = find_version("rgt", "__version__.py")

###################################################################################################
# Unsupported Platforms
###################################################################################################

supported_platforms = ["linux", "linux2", "darwin"]
if platform not in supported_platforms:
    print("ERROR: This package currently supports only unix-based systems (Linux and MAC OS X).")
    exit(0)

###################################################################################################
# Parameters
###################################################################################################

"""
Tools Dictionary:
  * Insert in the dictionary bellow a key+tuple X: (Y,Z,W,K) representing:
    - X: A string representing the name the user should provide for this script in order to install the tool.
    - Y: A string representing the name of the program which the user should type in the terminal in order to execute the tool.
    - Z: A string representing the path to the main function that executes the program.
    - W: A list with the package requirements for that program.
    - K: A list with external binary files that will be copied by the setup installation function to the user's bin folder.
Tools Dictionary Standard:
  * All programs should start with "rgt-" followed by the program name.
  * The main function called within the script must be termed "main".
"""

# TODO: sometime in the future, we should use environment markers for platform-specific dependencies.
# We currently can't, because the change is recent and many people have older python versions.
# We don't like crashes at installation time..

common_deps = ["cython",
               "numpy>=1.4.0",
               "scipy>=0.7.0",
               "pysam>=0.8.2",
               "pyBigWig",
               "PyVCF"]

if platform.startswith("darwin"):
    bin_dir = "mac"
    libRGT = "librgt_mac.so"
else:
    bin_dir = "linux"
    libRGT = "librgt_linux.so"
    common_deps.append("ngslib")

tools_dictionary = {
"core": (
    None,
    None,
    [],
    ["data/bin/"+bin_dir+"/bedToBigBed", "data/bin/"+bin_dir+"/bigBedToBed",
     "data/bin/"+bin_dir+"/wigToBigWig", "data/bin/"+bin_dir+"/bigWigMerge",
     "data/bin/"+bin_dir+"/bedGraphToBigWig"]
),
"motifanalysis": (
    "rgt-motifanalysis",
    "rgt.motifanalysis.Main:main",
    ["Biopython>=1.64", "fisher>=0.1.4"],
    ["data/bin/"+bin_dir+"/bedToBigBed", "data/bin/"+bin_dir+"/bigBedToBed"]
),
"hint": (
    "rgt-hint",
    "rgt.HINT.Main:main",
    ["scikit-learn>=0.14", "hmmlearn<0.2.0"],
    []
),
"THOR": (
    "rgt-THOR",
    "rgt.THOR.THOR:main",
    ["scikit-learn>=0.17.1", "hmmlearn<0.2.0", "matplotlib>=1.1.0", "mpmath", "HTSeq"],
    ["data/bin/"+bin_dir+"/wigToBigWig", "data/bin/"+bin_dir+"/bigWigMerge", "data/bin/"+bin_dir+"/bedGraphToBigWig"]
),
"filterVCF": (
    "rgt-filterVCF",
    "rgt.filterVCF.filterVCF:main",
    [],
    []
),
"viz": (
    "rgt-viz",
    "rgt.viz.Main:main",
    ["matplotlib>=1.1.0", "matplotlib_venn"],
    []
),
"TDF": (
    "rgt-TDF",
    "rgt.tdf.Main:main",
    ["matplotlib>=1.1.0", "natsort"],
    []
)
}

###################################################################################################
# Auxiliary Functions/Classes
###################################################################################################


# PassThroughOptionParser Class
class PassThroughOptionParser(OptionParser):
    """
    An unknown option pass-through implementation of OptionParser.
    When unknown arguments are encountered, bundle with largs and try again,
    until rargs is depleted.
    sys.exit(status) will still be called if a known argument is passed
    incorrectly (e.g. missing arguments or bad argument types, etc.)
    """
    def _process_args(self, largs, rargs, values):
        while rargs:
            try:
                OptionParser._process_args(self, largs, rargs, values)
            except (BadOptionError,AmbiguousOptionError), e:
                largs.append(e.opt_str)


# recursive_chown_chmod Function
def recursive_chown_chmod(path_to_walk, uid, gid, file_permission, path_permission):
    """
    Recursively applies chown from path.
    """
    for root_dir, directory_list, file_list in walk(path_to_walk):
        chown(root_dir, uid, gid)
        chmod(root_dir, path_permission)
        for f in file_list:
            current_complete_file = path.join(root_dir,f)
            chown(current_complete_file, uid, gid)
            chmod(current_complete_file, file_permission)

###################################################################################################
# Processing Input Arguments
###################################################################################################

# Parameters
rgt_data_base_name = "rgtdata"
usage_message = "python setup.py install [python options] [RGT options]"
version_message = "Regulatory Genomics Toolbox (RGT). Version: "+str(current_version)

# Initializing Option Parser
parser = PassThroughOptionParser(usage=usage_message, version=version_message)

# Parameter: RGT Data Location
param_rgt_data_location_name = "--rgt-data-path"
parser.add_option(param_rgt_data_location_name, type="string", metavar="STRING",
                  help="Path containing data used by RGT tool.",
                  dest="param_rgt_data_location", default=path.join(getenv('HOME'),rgt_data_base_name))

# Parameter: Tool
param_rgt_tool_name = "--rgt-tool"
parser.add_option(param_rgt_tool_name, type="string", metavar="STRING",
                  help=("The tool which will be installed. If this argument is not used, "
                        "then the complete package is installed. The current available options "
                        "are: "+", ".join(tools_dictionary.keys())+"; where 'core' means that "
                        "only the RGT python library will be installed with no further command-line "
                        "tools. You can also provide multiple tools in a list separated by comma."),
                  dest="param_rgt_tool", default=",".join(tools_dictionary.keys()))

# Processing Options
options, arguments = parser.parse_args()
if path.basename(options.param_rgt_data_location) != rgt_data_base_name:
    options.param_rgt_data_location = path.join(options.param_rgt_data_location,rgt_data_base_name)
if options.param_rgt_data_location[0] == "~":
    options.param_rgt_data_location = path.join(getenv('HOME'),options.param_rgt_data_location[2:])
options.param_rgt_tool = options.param_rgt_tool.split(",")

# Manually Removing Additional Options from sys.argv
new_sys_argv = []
for e in sys.argv:
    if param_rgt_data_location_name == e[:len(param_rgt_data_location_name)]:
        continue
    elif param_rgt_tool_name == e[:len(param_rgt_tool_name)]:
        continue
    new_sys_argv.append(e)

sys.argv = new_sys_argv

# Defining entry points
current_entry_points = {"console_scripts" : []}
for tool_option in options.param_rgt_tool:
    if tool_option != "core":
        current_entry_points["console_scripts"].append(" = ".join(tools_dictionary[tool_option][:2]))

# Defining install requirements
current_install_requires = common_deps
for tool_option in options.param_rgt_tool:
    current_install_requires += tools_dictionary[tool_option][2]

###################################################################################################
# Creating Data Path
###################################################################################################

# Creating Data Path
if not path.exists(options.param_rgt_data_location):
    makedirs(options.param_rgt_data_location)

# Creating data.config
data_config_file_name = path.join(options.param_rgt_data_location, "data.config")
# if not os.path.isfile(data_config_file_name):
data_config_file = open(data_config_file_name,"w")
data_config_file.write("# Configuration file loaded at rgt startup. CAREFUL: any changes shall be overwritten\n"
                       "# whenever rgt is (re)installed. Use data.config.user for permanent changes.\n\n")

genome = "mm9"
genome_dir = path.join(options.param_rgt_data_location, genome)
data_config_file.write("["+genome+"]\n")
data_config_file.write("genome: "+path.join(genome_dir,"genome_mm9.fa\n"))
data_config_file.write("chromosome_sizes: "+path.join(genome_dir,"chrom.sizes.mm9\n"))
data_config_file.write("genes_Gencode: "+path.join(genome_dir,"genes_Gencode_mm9.bed\n"))
data_config_file.write("genes_RefSeq: "+path.join(genome_dir,"genes_RefSeq_mm9.bed\n"))
data_config_file.write("annotation: "+path.join(genome_dir,"gencode.vM1.annotation.gtf\n"))
data_config_file.write("gene_alias: "+path.join(genome_dir,"alias_mouse.txt\n\n"))
genome = "mm10"
genome_dir = path.join(options.param_rgt_data_location, genome)
data_config_file.write("["+genome+"]\n")
data_config_file.write("genome: "+path.join(genome_dir,"genome_mm10.fa\n"))
data_config_file.write("chromosome_sizes: "+path.join(genome_dir,"chrom.sizes.mm10\n"))
data_config_file.write("genes_Gencode: "+path.join(genome_dir,"genes_Gencode_mm10.bed\n"))
data_config_file.write("genes_RefSeq: "+path.join(genome_dir,"genes_RefSeq_mm10.bed\n"))
data_config_file.write("annotation: "+path.join(genome_dir,"gencode.vM11.annotation.gtf\n"))
data_config_file.write("gene_alias: "+path.join(genome_dir,"alias_mouse.txt\n\n"))
genome = "hg19"
genome_dir = path.join(options.param_rgt_data_location, genome)
data_config_file.write("["+genome+"]\n")
data_config_file.write("genome: "+path.join(genome_dir,"genome_hg19.fa\n"))
data_config_file.write("chromosome_sizes: "+path.join(genome_dir,"chrom.sizes.hg19\n"))
data_config_file.write("genes_Gencode: "+path.join(genome_dir,"genes_Gencode_hg19.bed\n"))
data_config_file.write("genes_RefSeq: "+path.join(genome_dir,"genes_RefSeq_hg19.bed\n"))
data_config_file.write("annotation: "+path.join(genome_dir,"gencode.v19.annotation.gtf\n"))
data_config_file.write("gene_alias: "+path.join(genome_dir,"alias_human.txt\n\n"))
genome = "hg38"
genome_dir = path.join(options.param_rgt_data_location, genome)
data_config_file.write("["+genome+"]\n")
data_config_file.write("genome: "+path.join(genome_dir,"genome_hg38.fa\n"))
data_config_file.write("chromosome_sizes: "+path.join(genome_dir,"chrom.sizes.hg38\n"))
data_config_file.write("genes_Gencode: "+path.join(genome_dir,"genes_Gencode_hg38.bed\n"))
data_config_file.write("genes_RefSeq: "+path.join(genome_dir,"genes_RefSeq_hg38.bed\n"))
data_config_file.write("annotation: "+path.join(genome_dir,"gencode.v24.annotation.gtf\n"))
data_config_file.write("gene_alias: "+path.join(genome_dir,"alias_human.txt\n\n"))
genome = "zv9"
genome_dir = path.join(options.param_rgt_data_location, genome)
data_config_file.write("["+genome+"]\n")
data_config_file.write("genome: "+path.join(genome_dir,"genome_zv9_ensembl_release_79.fa\n"))
data_config_file.write("chromosome_sizes: "+path.join(genome_dir,"chrom.sizes.zv9\n"))
data_config_file.write("gene_regions: "+path.join(genome_dir,"genes_zv9.bed\n"))
data_config_file.write("annotation: "+path.join(genome_dir,"Danio_rerio.Zv9.79.gtf\n"))
data_config_file.write("gene_alias: "+path.join(genome_dir,"alias_zebrafish.txt\n\n"))
genome = "zv10"
genome_dir = path.join(options.param_rgt_data_location, genome)
data_config_file.write("["+genome+"]\n")
data_config_file.write("genome: "+path.join(genome_dir,"genome_zv10_ensembl_release_84.fa\n"))
data_config_file.write("chromosome_sizes: "+path.join(genome_dir,"chrom.sizes.zv10\n"))
data_config_file.write("gene_regions: "+path.join(genome_dir,"genes_zv10.bed\n"))
data_config_file.write("annotation: "+path.join(genome_dir,"Danio_rerio.GRCz10.84.gtf\n"))
data_config_file.write("gene_alias: "+path.join(genome_dir,"alias_zebrafish.txt\n\n"))

data_config_file.write("[MotifData]\n")
data_config_file.write("pwm_dataset: motifs\n")
data_config_file.write("logo_dataset: logos\n")
data_config_file.write("repositories: jaspar_vertebrates,uniprobe_primary\n\n")
data_config_file.write("[HmmData]\n")
data_config_file.write("default_hmm_dnase: fp_hmms/dnase.hmm\n")
data_config_file.write("default_hmm_dnase_bc: fp_hmms/dnase_bc.hmm\n")
data_config_file.write("default_hmm_atac: fp_hmms/atac.hmm\n")
data_config_file.write("default_hmm_atac_bc: fp_hmms/atac_bc.hmm\n")
data_config_file.write("default_hmm_histone: fp_hmms/histone.hmm\n")
data_config_file.write("default_hmm_dnase_histone: fp_hmms/dnase_histone.hmm\n")
data_config_file.write("default_hmm_dnase_histone_bc: fp_hmms/dnase_histone_bc.hmm\n")
data_config_file.write("default_hmm_atac_histone: fp_hmms/atac_histone.hmm\n")
data_config_file.write("default_hmm_atac_histone_bc: fp_hmms/atac_histone_bc.hmm\n")
data_config_file.write("default_bias_table_F_SH: fp_hmms/single_hit_bias_table_F.txt\n")
data_config_file.write("default_bias_table_R_SH: fp_hmms/single_hit_bias_table_R.txt\n")
data_config_file.write("default_bias_table_F_DH: fp_hmms/double_hit_bias_table_F.txt\n")
data_config_file.write("default_bias_table_R_DH: fp_hmms/double_hit_bias_table_R.txt\n")
data_config_file.write("default_bias_table_F_ATAC: fp_hmms/atac_bias_table_F.txt\n")
data_config_file.write("default_bias_table_R_ATAC: fp_hmms/atac_bias_table_R.txt\n\n")
data_config_file.write("[Library]\n")
data_config_file.write("path_triplexator: " + path.join(options.param_rgt_data_location, "lib/libtriplexator.so") + "\n")
data_config_file.write("path_c_rgt: " + path.join(options.param_rgt_data_location, "lib/"+libRGT) + "\n")

data_config_file.close()

# Creating data.config.user, but only if not already present
user_config_file_name = path.join(options.param_rgt_data_location, "data.config.user")
if not os.path.isfile(user_config_file_name):
    user_config_file = open(user_config_file_name, "w")

    user_config_file.write("# Here you can overwrite any property set in the data.config file. It shall not be\n"
                           "# be overwritten in any case, so if you are experiencing problems rename or remove this\n"
                           "# file. See data.config for how the file should be formatted.\n\n")
    genome = "self_defined"
    genome_dir = path.join(options.param_rgt_data_location, genome)
    user_config_file.write("# Template to add a genomic section.\n")
    user_config_file.write("#["+genome+"]\n")
    user_config_file.write("#genome: undefined\n")
    user_config_file.write("#chromosome_sizes: undefined\n")
    user_config_file.write("#gene_regions: undefined\n")
    user_config_file.write("#annotation: undefined\n")
    user_config_file.write("#gene_alias: undefined\n\n")

# Creating data.config.path
script_dir = path.dirname(path.abspath(__file__))
data_config_path_file_name = path.join(script_dir, "rgt", "data.config.path")
data_config_path_file = open(data_config_path_file_name, "w")
data_config_path_file.write(data_config_file_name)
data_config_path_file.close()

# Copying data from package folder to installation folder
"""
Copy Files Dictionary:
  * Insert in the dictionary below a key + list in the format X:[Y1,Y2,...Yn], representing:
    - X: A path in the data folder structure.
    - Y1,Y2,...Yn: files/folders inside the path X to be copied.
"""
copy_files_dictionary = {
".": ["setupGenomicData.py","setupLogoData.py"],
"lib": ["libtriplexator.so", libRGT],
"hg19": ["genes_Gencode_hg19.bed","chrom.sizes.hg19","alias_human.txt","genes_RefSeq_hg19.bed"],
"hg38": ["genes_Gencode_hg38.bed","chrom.sizes.hg38","alias_human.txt","genes_RefSeq_hg38.bed"],
"mm9": ["genes_Gencode_mm9.bed","chrom.sizes.mm9","alias_mouse.txt","genes_RefSeq_mm9.bed"],
"mm10": ["genes_Gencode_mm10.bed","chrom.sizes.mm10","alias_mouse.txt","genes_RefSeq_mm10.bed"],
"zv9": ["genes_zv9.bed","chrom.sizes.zv9","alias_zebrafish.txt"],
"zv10": ["genes_zv10.bed","chrom.sizes.zv10","alias_zebrafish.txt"],
"fp_hmms": ["dnase.hmm", "dnase_bc.hmm", "histone.hmm", "dnase_histone.hmm", "dnase_histone_bc.hmm", "single_hit_bias_table_F.txt", "single_hit_bias_table_R.txt", "atac.hmm", "atac_bc.hmm", "atac_bias_table_F.txt", "atac_bias_table_R.txt", "atac_histone.hmm", "atac_histone_bc.hmm", "double_hit_bias_table_F.txt", "double_hit_bias_table_R.txt", "H3K4me3_proximal.hmm"],
"motifs": ["jaspar_vertebrates", "uniprobe_primary", "uniprobe_secondary", "hocomoco", "hocomoco.fpr", "jaspar_vertebrates.fpr", "uniprobe_primary.fpr", "uniprobe_secondary.fpr"],
"fig": ["rgt_logo.gif","style.css","default_motif_logo.png","jquery-1.11.1.js","jquery.tablesorter.min.js","tdf_logo.png", "viz_logo.png"],
}
for copy_folder in copy_files_dictionary.keys():
    copy_dest_path = path.join(options.param_rgt_data_location, copy_folder)
    if not path.exists(copy_dest_path): makedirs(copy_dest_path)
    for copy_file in copy_files_dictionary[copy_folder]:
        copy_source_file = path.join(script_dir,"data", copy_folder, copy_file)
        copy_dest_file = path.join(copy_dest_path,copy_file)
        if os.path.isfile(copy_source_file):
            copy(copy_source_file, copy_dest_file)
        else:
            dir_util.copy_tree(copy_source_file, copy_dest_file)
        
        # if not path.exists(copy_dest_file): 
        # try: dir_util.copy_tree(copy_source_file, copy_dest_file)
        # except OSError as exc:
        #     if exc.errno == ENOTDIR: 
        #         copy(copy_source_file, copy_dest_file)
        #     else:
        #         raise
    
###################################################################################################
# Setup Function
###################################################################################################

# Parameters
short_description = "Toolkit to perform regulatory genomics data analysis"
classifiers_list = ["Topic :: Scientific/Engineering :: Bio-Informatics",
                    "Topic :: Scientific/Engineering :: Artificial Intelligence"]
keywords_list = ["ChIP-seq", "DNase-seq", "Peak Calling", "Motif Discovery", "Motif Enrichment", "HMM"]
author_list = ["Eduardo G. Gusmao", "Manuel Allhoff", "Joseph Chao-Chung Kuo", "Fabio Ticconi", "Ivan G. Costa"]
corresponding_mail = "software@costalab.org"
license_type = "GPL"
package_data_dictionary = {"rgt": [path.basename(data_config_path_file_name)]}

# External scripts
external_scripts = []
for tool_option in options.param_rgt_tool:
    for e in tools_dictionary[tool_option][3]:
        external_scripts.append(e)

# Fetching Additional Structural Files
readme_file_name = path.join(path.dirname(path.abspath(__file__)), "README.rst")

# Fetching Long Description
readme_file = open(readme_file_name, "r")
long_description = readme_file.read()
readme_file.close()

# Setup Function

setup(name="RGT",
      version=current_version,
      description=short_description,
      long_description=long_description,
      classifiers=classifiers_list,
      keywords=", ".join(keywords_list),
      author=", ".join(author_list),
      author_email=corresponding_mail,
      license=license_type,
      packages=find_packages(),
      package_data=package_data_dictionary,
      entry_points=current_entry_points,
      install_requires=current_install_requires,
      scripts=external_scripts,
      url="http://www.regulatory-genomics.org",
      download_url="https://github.com/CostaLab/reg-gen/archive/{0}.zip".format(current_version),
      platforms=supported_platforms)

###################################################################################################
# Termination
###################################################################################################

# Removing data.config.path
remove(data_config_path_file_name)

# Modifying Permissions when Running Superuser/Admin
# $SUDO_USER exists only if you are sudo, and returns the original user name
current_user = getenv("SUDO_USER")
default_file_permission = 0644
default_path_permission = 0755

if current_user:
    current_user_uid = getpwnam(current_user).pw_uid
    current_user_gid = getpwnam(current_user).pw_gid
    recursive_chown_chmod(options.param_rgt_data_location,current_user_uid,current_user_gid,default_file_permission,default_path_permission)



"""
Copyright (C) 2014-2016 Manuel Allhoff (allhoff@aices.rwth-aachen.de)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

@author: Manuel Allhoff
"""

from __future__ import print_function
import sys
import numpy as np
from os import path
from random import sample
from rgt.CoverageSet import CoverageSet
from rgt.CoverageSet import get_gc_context
from normalize import get_normalization_factor

EPSILON=1e-320

class DualCoverageSet():
    def _get_BAM_names(self, input, ip):
        # get names of BAM files for bw files
        name_bam = path.splitext(path.basename(ip))[0]
        if input is not None:
            name_input = path.splitext(path.basename(input))[0]
        else:
            name_input = None

        return name_bam, name_input

    def __init__(self, name, region, genome_path, binsize, stepsize, rmdup, file_1, ext_1, file_2, ext_2, \
                 input_1, ext_input_1, input_factor_1, input_2, ext_input_2, input_factor_2, chrom_sizes, verbose,
                 norm_strategy, no_gc_content, deadzones, \
                 factor_input_1, factor_input_2, chrom_sizes_dict, debug, tracker):
        self.genomicRegions = region
        self.binsize = binsize
        self.stepsize = stepsize
        self.name = name
        self.cov1 = CoverageSet('first file', region)
        self.cov2 = CoverageSet('second file', region)

        print("Loading reads...", file=sys.stderr)
        self.cov1.coverage_from_bam(bam_file=file_1, extension_size=ext_1, rmdup=rmdup, binsize=binsize,
                                    stepsize=stepsize, mask_file=deadzones)
        self.cov2.coverage_from_bam(bam_file=file_2, extension_size=ext_2, rmdup=rmdup, binsize=binsize,
                                    stepsize=stepsize, mask_file=deadzones)

        map_input = {1: {'input': input_1, 'input_factor': input_factor_1, 'ext': ext_input_1, 'cov-ip': self.cov1,
                         'ip': file_1},
                     2: {'input': input_2, 'input_factor': input_factor_2, 'ext': ext_input_2, 'cov-ip': self.cov2,
                         'ip': file_2}}

        if not no_gc_content and input_1 is not None and input_2 is not None:
            print("Computing GC content", file=sys.stderr)
        else:
            print("Do not compute GC content", file=sys.stderr)

        for i in [1, 2]:
            input = map_input[i]
            name_bam, name_input = self._get_BAM_names(input['input'], input['ip'])

            if debug:  # 0: output raw IP
                input['cov-ip'].write_bigwig(name + '-debug-0-' + name_bam + '.bw', chrom_sizes)

            if input['input'] is not None:
                input['cov-input'] = CoverageSet('%s file' % input['input'], region)
                input['cov-input'].coverage_from_bam(bam_file=input['input'], extension_size=input['ext'], rmdup=rmdup,
                                                     binsize=binsize, stepsize=stepsize)
                map_input[i]['cov-input'] = input['cov-input']

            if not no_gc_content and input['input'] is not None:
                gc_content_cov, avg_gc_content, gc_hist = get_gc_context(stepsize, binsize, genome_path,
                                                                         input['cov-input'].coverage, chrom_sizes_dict)

                self._norm_gc_content(input['cov-ip'].coverage, gc_content_cov, avg_gc_content)
                self._norm_gc_content(input['cov-input'].coverage, gc_content_cov, avg_gc_content)

                if debug:  # 1: output after GC
                    self.print_gc_hist(name + '-' + name_input, gc_hist)  # print hist data
                    input['cov-input'].write_bigwig(name + '-debug-1-' + name_input + '.bw', chrom_sizes)
                    input['cov-ip'].write_bigwig(name + '-debug-1-' + name_bam + '.bw', chrom_sizes)

        norm_done = False
        print("Normalizing signals", file=sys.stderr)
        for i in [1, 2]:
            input = map_input[i]
            name_bam, name_input = self._get_BAM_names(input['input'], input['ip'])

            # TODO: uncomment here!
            norm_done = self.normalization(map_input, i, norm_strategy, norm_done, name, debug, factor_input_1,
                                           factor_input_2, chrom_sizes_dict, tracker)

            if input['input'] is not None:
                input['cov-input'].write_bigwig(name + '-' + name_input + '-normalized.bw', chrom_sizes)
            input['cov-ip'].write_bigwig(name + '-' + name_bam + '-normalized.bw', chrom_sizes)

        # make one array for the coverage
        self.first_overall_coverage = reduce(lambda x, y: np.concatenate((x, y)),
                                             [self.cov1.coverage[i] for i in range(len(self.cov1.genomicRegions))])
        self.second_overall_coverage = reduce(lambda x, y: np.concatenate((x, y)),
                                              [self.cov2.coverage[i] for i in range(len(self.cov2.genomicRegions))])
        assert (len(self.first_overall_coverage) == len(self.second_overall_coverage))

        self.scores = np.zeros(len(self.first_overall_coverage))
        self.indices_of_interest = []

    def normalization(self, map_input, i, norm_strategy, norm_done, name, debug, factor_input_1, factor_input_2,
                      chrom_sizes_dict, tracker):
        input = map_input[i]

        # compute normalization factor
        # pre-defined values
        if input['input_factor'] is not None and i != 1:
            print("Normalize by Diaz and pre-defined values...", input['input_factor'], file=sys.stderr)
            print("Normalize file 1 with input normalization factor %s" % (map_input[1]['input_factor']),
                  file=sys.stderr)
            print("Normalize file 2 with input normalization factor %s" % (map_input[2]['input_factor']),
                  file=sys.stderr)
            tracker.write(text=str(map_input[1]['input_factor']) + ',' + str(map_input[2]['input_factor']),
                          header="Predefined Normalization factor of Input")

            map_input[1]['cov-input'].scale(map_input[1]['input_factor'])
            map_input[2]['cov-input'].scale(map_input[2]['input_factor'])
            map_input[1]['cov-ip'].subtract(map_input[1]['cov-input'])
            map_input[2]['cov-ip'].subtract(map_input[2]['cov-input'])

        # naive norm.
        if not norm_done and norm_strategy == 1:
            if factor_input_1 is None or factor_input_2 is None:
                s1 = sum([sum(map_input[1]['cov-ip'].coverage[i]) for i in
                          range(len(map_input[1]['cov-ip'].genomicRegions))])
                s2 = sum([sum(map_input[2]['cov-ip'].coverage[i]) for i in
                          range(len(map_input[2]['cov-ip'].genomicRegions))])
                if s1 > s2:
                    map_input[2]['cov-ip'].scale(s1 / float(s2))
                    print("Normalize file 2 by signal with estimated factor %s " % (round(s1 / float(s2), 3)),
                          file=sys.stderr)
                    tracker.write(text=str(round(s1 / float(s2), 3)), header="Normalization factor of signal 2")
                elif s2 >= s1:
                    print("Normalize file 1 by signal with estimated factor %s " % (round(s2 / float(s1), 3)),
                          file=sys.stderr)
                    tracker.write(text=str(round(s2 / float(s1), 3)), header="Normalization factor of signal 1")
                    map_input[1]['cov-ip'].scale(s2 / float(s1))

                norm_done = True
            else:
                map_input[1]['cov-ip'].scale(factor_input_1)
                print("Normalize file 1 by signal with given factor %s " % round(factor_input_1, 3), file=sys.stderr)
                tracker.write(text=str(round(factor_input_1, 3)), header="Predefined Normalization factor of signal 1")

                map_input[2]['cov-ip'].scale(factor_input_2)
                print("Normalize file 2 by signal with given factor %s " % round(factor_input_2, 3), file=sys.stderr)
                tracker.write(text=str(round(factor_input_2, 3)), header="Predefined Normalization factor of signal 2")
                norm_done = True

        # diaz and naive
        if i != 1 and norm_strategy == 5:
            # apply diaz
            _, map_input[1]['input_factor'] = get_normalization_factor(map_input[1]['ip'], map_input[1]['input'],
                                                                       step_width=1000, zero_counts=0, \
                                                                       filename=name + '-norm' + str(i), debug=debug,
                                                                       chrom_sizes_dict=chrom_sizes_dict,
                                                                       two_sample=False)
            _, map_input[2]['input_factor'] = get_normalization_factor(map_input[2]['ip'], map_input[2]['input'],
                                                                       step_width=1000, zero_counts=0, \
                                                                       filename=name + '-norm' + str(i), debug=debug,
                                                                       chrom_sizes_dict=chrom_sizes_dict,
                                                                       two_sample=False)

            print("Normalize input with factor %s and %s" % (
            round(map_input[1]['input_factor'], 3), round(map_input[2]['input_factor'], 3)), file=sys.stderr)
            tracker.write(
                text=str(round(map_input[1]['input_factor'], 3)) + ',' + str(round(map_input[2]['input_factor'], 3)),
                header="Input Normalization factors")

            map_input[1]['cov-input'].scale(map_input[1]['input_factor'])
            map_input[2]['cov-input'].scale(map_input[2]['input_factor'])

            map_input[1]['cov-ip'].subtract(map_input[1]['cov-input'])
            map_input[2]['cov-ip'].subtract(map_input[2]['cov-input'])

            if factor_input_1 is None or factor_input_2 is None:
                # apply naive method
                s1 = sum([sum(map_input[1]['cov-ip'].coverage[i]) for i in
                          range(len(map_input[1]['cov-ip'].genomicRegions))])
                s2 = sum([sum(map_input[2]['cov-ip'].coverage[i]) for i in
                          range(len(map_input[2]['cov-ip'].genomicRegions))])

                if s1 > s2:
                    map_input[2]['cov-ip'].scale(s1 / float(s2))
                    print("Normalize file 2 by signal with estimated factor %s " % (round(s1 / float(s2), 3)),
                          file=sys.stderr)
                    tracker.write(text=str(round(s1 / float(s2), 3)), header="Normalization factor of signal 2")
                elif s2 >= s1:
                    print("Normalize file 1 by signal with estimated factor %s " % (round(s2 / float(s1), 3)),
                          file=sys.stderr)
                    map_input[1]['cov-ip'].scale(s2 / float(s1))
                    tracker.write(text=str(round(s2 / float(s1), 3)), header="Normalization factor of signal 1")
            else:
                map_input[1]['cov-ip'].scale(factor_input_1)
                print("Normalize file 1 by signal with given factor %s " % round(factor_input_1, 3), file=sys.stderr)
                tracker.write(text=str(round(factor_input_1, 3)), header="Normalization factor of signal 1")
                map_input[2]['cov-ip'].scale(factor_input_2)
                print("Normalize file 2 by signal with given factor %s " % round(factor_input_2, 3), file=sys.stderr)
                tracker.write(text=str(round(factor_input_2, 3)), header="Normalization factor of signal 2")
        return norm_done

    def print_gc_hist(self, name, gc_hist):
        f = open(name + '-gc-content.data', 'w')
        for i in range(len(gc_hist)):
            print(i, gc_hist[i], file=f)
        f.close()

    def _norm_gc_content(self, cov, gc_cov, gc_avg):
        for i in range(len(cov)):
            assert len(cov[i]) == len(gc_cov[i])
            #            cov[i] = gc_cov[i]
            cov[i] = np.array(cov[i])
            gc_cov[i] = np.array(gc_cov[i])
            gc_cov[i][gc_cov[i] < EPSILON] = gc_avg  # sometimes zeros occur, do not consider
            cov[i] = cov[i] * gc_avg / gc_cov[i]
            cov[i] = cov[i].clip(0, max(max(cov[i]), 0))  # neg. values to 0
            cov[i] = cov[i].astype(int)

    def _index2coordinates(self, index):
        """Translate index within coverage array to genomic coordinates."""
        iter = self.genomicRegions.__iter__()
        r = iter.next()
        sum = r.final
        last = 0
        i = 0
        while sum <= index * self.stepsize:
            last += len(self.cov1.coverage[i])
            try:
                r = iter.next()
            except StopIteration:
                sum += r.final
                i += 1
                break
            sum += r.final
            i += 1

        return r.chrom, (index - last) * self.stepsize, \
               min((index - last) * self.stepsize + self.stepsize, r.final)

    def __len__(self):
        """Return number of observations."""
        return len(self.indices_of_interest)

    def get_observation(self, mask=np.array([])):
        """Return indices of observations. Do not consider indices contained in <mask> array"""
        if not mask.size:
            mask = np.array([True] * len(self.first_overall_coverage))
        return np.array([self.first_overall_coverage[mask], self.second_overall_coverage[mask]]).T

    def _compute_score(self):
        """Compute score for each observation (based on Xu et al.)"""
        self.scores = self.first_overall_coverage / float(sum(self.first_overall_coverage)) + \
                      self.second_overall_coverage / float(sum(self.second_overall_coverage))

    def compute_putative_region_index(self, l=5):
        """Compute putative differential peak regions as follows: 
        - score must be > 2/(m*n) (m=#obs, n=0.9 (default) )
        - overall coverage in library 1 and 2 must be > 3
        - extend resulting sites by l steps in both directions. """
        m = len(self.first_overall_coverage)
        n = 0.9
        self._compute_score()
        #        print('before filter step:', len(self.scores), file=sys.stderr)
        # print(self.first_overall_coverage, self.second_overall_coverage, file=sys.stderr)
        self.indices_of_interest = np.where(self.scores > 2 / (m * n))[0]
        #        print('after first filter step: ', len(self.indices_of_interest), file=sys.stderr)
        tmp = np.where(self.first_overall_coverage + self.second_overall_coverage > 3)[0]
        tmp2 = np.intersect1d(self.indices_of_interest, tmp)
        #        print('length of intersection set: ', len(tmp), file=sys.stderr)
        self.indices_of_interest = tmp2
        #        print('after second filter step: ', len(self.indices_of_interest), file=sys.stderr)
        # extend regions by l steps
        tmp = set()
        for i in self.indices_of_interest:
            for j in range(max(0, i - l), i + l + 1):
                tmp.add(j)
        tmp = list(tmp)
        tmp.sort()
        self.indices_of_interest = np.array(tmp)

    def get_initial_dist(self, filename):
        """Write BED file with initial state distribution"""
        states = []
        threshold = 2.0
        for i in self.indices_of_interest:
            c1 = self.first_overall_coverage[i]
            c2 = self.second_overall_coverage[i]

            if c1 + c2 <= 3:
                state = 0
            elif c1 / max(float(c2), 1) > threshold or c1 - c2 > 10:
                state = 1
            elif c1 / max(float(c2), 1) < 1 / threshold or c2 - c1 > 10:
                state = 2
            else:
                state = 0

            states.append(state)

        f = open(filename, 'w')
        for j in range(len(states)):
            i = self.indices_of_interest[j]
            chrom, start, end = self._index2coordinates(i)
            s = states[j]
            print(chrom, start, end, s, self.first_overall_coverage[i], self.second_overall_coverage[i], sep='\t',
                  file=f)

        f.close()

    def write_putative_regions(self, path):
        """Write putative regions (defined by criteria mentioned in method) as BED file."""
        with open(path, 'w') as f:
            for i in self.indices_of_interest:
                chrom, start, end = self._index2coordinates(i)
                print(chrom, start, end, file=f)

    def get_training_set(self, exp_data, x, verbose, name, debug, constraint_chrom, min_fc=2.0):
        """Return linked genomic positions (at least <x> positions) to train HMM.
        Grep randomly a position within a putative region, and take then the entire region."""
        training_set = set()
        ts1 = set()
        ts2 = set()
        threshold = min_fc
        diff_cov = 10

        if constraint_chrom is not None:
            print("HMM training set based on %s" % constraint_chrom, file=sys.stderr)

        for i in range(len(self.indices_of_interest)):
            chrom, start, end = self._index2coordinates(i)
            if constraint_chrom is not None and chrom != constraint_chrom:
                continue
            cov1 = exp_data.first_overall_coverage[self.indices_of_interest[i]]
            cov2 = exp_data.second_overall_coverage[self.indices_of_interest[i]]

            if cov1 / max(float(cov2), 1) > threshold or cov1 - cov2 > diff_cov:
                ts1.add(i)
            if cov1 / max(float(cov2), 1) < 1 / threshold or cov2 - cov1 > diff_cov:
                ts2.add(i)

        l = min(min(len(ts1), len(ts2)), x)
        tmp = set(sample(ts1, l)) | set(sample(ts2, l))

        for i in tmp:
            training_set.add(self.indices_of_interest[i])
            # search up
            while i + 1 < len(self.indices_of_interest) and self.indices_of_interest[i + 1] == self.indices_of_interest[
                i] + 1:
                training_set.add(self.indices_of_interest[i + 1])
                i += 1
            # search down
            while i - 1 > 0 and self.indices_of_interest[i - 1] == self.indices_of_interest[i] - 1:
                training_set.add(self.indices_of_interest[i - 1])
                i -= 1


                #        while len(training_set) < x:
                #            i = randrange(1, len(self.indices_of_interest)-1)
                #
                #            if i in used:
                #                continue #todo: super ugly
                #            used.add(i)
                #            training_set.add(self.indices_of_interest[i])
                #            #search up
                #            while i+1 < len(self.indices_of_interest) and self.indices_of_interest[i+1] == self.indices_of_interest[i]+1:
                #                training_set.add(self.indices_of_interest[i+1])
                #                i += 1
                #            #search down
                #            while i-1 > 0 and self.indices_of_interest[i-1] == self.indices_of_interest[i]-1:
                #                training_set.add(self.indices_of_interest[i-1])
                #                i -= 1

        training_set = list(training_set)
        training_set.sort()
        if debug:
            f = open(name + '-trainingset.bed', 'w')
            for l in training_set:
                chrom, s, e = self._index2coordinates(l)
                print(chrom, s, e, sep='\t', file=f)
            f.close()

        return np.array(training_set)




###################################################################################################
# Setup Binaries Script
#   This script will fetch the necessary binaries for all tools. Be aware that it will consider
#   only 64 bits system (linux and MAC OS X).
###################################################################################################

# Import
from os import system, path, walk, remove, mkdir
from sys import platform
from glob import glob
from optparse import OptionParser

# Optional Input
usage_message = "python setupData.py [options]"

# Initializing Option Parser
parser = OptionParser(usage = usage_message)
options, arguments = parser.parse_args()


###################################################################################################
# Parameters
###################################################################################################

# Current rgt data path
curr_dir = path.dirname(path.realpath(__file__))

# Platform
supported_platforms = ["linux","linux2","darwin"]
if(platform not in supported_platforms):
    print("ERROR: This package currently supports only unix-based systems (Linux and MAC OS X).")
    exit(0)

###################################################################################################
# Binary Data
###################################################################################################

# Fetching UCSC binaries
ucsc_util_list = ["bigBedToBed","bedToBigBed"]
if("linux" in platform): bin_root = "http://hgdownload.cse.ucsc.edu/admin/exe/linux.x86_64/"
else: bin_root = "http://hgdownload.cse.ucsc.edu/admin/exe/macOSX.x86_64/"
out_loc = path.join(curr_dir,"bin")
if(not path.exists(out_loc)): mkdir(out_loc)
for util_tool in ucsc_util_list:
    print "Downloading UCSC binary script "+util_tool
    if(not path.isfile(path.join(out_loc,util_tool))):
        system("wget "+bin_root+util_tool+" -P "+out_loc)
    print "OK"





###################################################################################################
# Libraries
###################################################################################################

# Python
import warnings

warnings.filterwarnings("ignore")
from math import log, ceil, floor, isnan

# Internal
from ..Util import ErrorHandler
from ..Util import AuxiliaryFunctions
from pileupRegion import PileupRegion

# External
from pysam import __version__ as ps_version
from pysam import Samfile
from pysam import Fastafile
from numpy import exp, array, abs, int, mat, linalg, convolve, nan, nan_to_num
from scipy.stats import scoreatpercentile

"""
Processes DNase-seq and histone modification signal for
HMM footprinting input.

Authors: Eduardo G. Gusmao.
"""


class GenomicSignal:
    """
    Represents a genomic signal. It should be used to fetch normalized and slope
    signals from a bam file.
    Usage:
    1. Initialize class.
    2. Call load_sg_coefs once.
    3. Call get_signal as many times as needed.

    Authors: Eduardo G. Gusmao.
    """

    def __init__(self, file_name):
        """ 
        Initializes GenomicSignal.
        """
        self.file_name = file_name
        self.sg_coefs = None
        self.bam = Samfile(file_name, "rb")

    def load_sg_coefs(self, slope_window_size):
        """ 
        Loads Savitzky-Golay coefficients into self.sg_coefs based on a slope_window_size.

        Keyword arguments:
        slope_window_size -- Window size of Savitzky-Golay coefficients.

        Return:
        None -- It updates self.sg_coefs.
        """
        self.sg_coefs = self.savitzky_golay_coefficients(slope_window_size, 2, 1)

    def get_tag_count(self, ref, start, end, downstream_ext, upstream_ext, forward_shift, reverse_shift,
                      initial_clip=1000):
        """
        Gets the tag count associated with self.bam based on start, end and ext.

        Keyword arguments:
        ref -- Chromosome name.
        start -- Initial genomic coordinate of signal.
        end -- Final genomic coordinate of signal.
        downstream_ext -- Number of bps to extend towards the downstream region (right for forward strand and left for reverse strand).
        upstream_ext -- Number of bps to extend towards the upstream region (left for forward strand and right for reverse strand).
        forward_shift -- Number of bps to shift the reads aligned to the forward strand. Can be a positive number for a shift towards the downstream region (towards the inside of the aligned read) and a negative number for a shift towards the upstream region.
        reverse_shift -- Number of bps to shift the reads aligned to the reverse strand. Can be a positive number for a shift towards the upstream region and a negative number for a shift towards the downstream region (towards the inside of the aligned read).
        initial_clip -- Signal will be initially clipped at this level to avoid outliers.

        Return:
        tag_count -- Total signal.
        """

        # Fetch raw signal
        pileup_region = PileupRegion(start, end, downstream_ext, upstream_ext, forward_shift, reverse_shift)
        if (ps_version == "0.7.5"):
            self.bam.fetch(reference=ref, start=start, end=end, callback=pileup_region)
        else:
            iter = self.bam.fetch(reference=ref, start=start, end=end)
            for alignment in iter: pileup_region.__call__(alignment)
        raw_signal = array([min(e, initial_clip) for e in pileup_region.vector])

        # Std-based clipping
        mean = raw_signal.mean()
        std = raw_signal.std()
        clip_signal = [min(e, mean + (10 * std)) for e in raw_signal]

        # Tag count
        try:
            tag_count = sum(clip_signal)
        except Exception:
            tag_count = 0

        return tag_count

    def get_signal(self, ref, start, end, downstream_ext, upstream_ext, forward_shift, reverse_shift,
                   initial_clip=1000, per_norm=98, per_slope=98,
                   bias_table=None, genome_file_name=None, print_raw_signal=False,
                   print_bc_signal=False, print_norm_signal=False, print_slope_signal=False,
                   strands_specific=False):
        """
        Gets the signal associated with self.bam based on start, end and ext.
        initial_clip, per_norm and per_slope are used as normalization factors during the normalization
        and slope evaluation procedures.

        Keyword arguments:
        ref -- Chromosome name.
        start -- Initial genomic coordinate of signal.
        end -- Final genomic coordinate of signal.
        initial_clip -- Signal will be initially clipped at this level to avoid outliers.
        per_norm -- Percentile value for 'hon_norm' function of the normalized signal.
        per_slope -- Percentile value for 'hon_norm' function of the slope signal.
        bias_table -- Bias table to perform bias correction.
        genome_file_name -- Genome to perform bias correction.
        downstream_ext -- Number of bps to extend towards the downstream region
        (right for forward strand and left for reverse strand).
        upstream_ext -- Number of bps to extend towards the upstream region
        (left for forward strand and right for reverse strand).
        forward_shift -- Number of bps to shift the reads aligned to the forward strand.
        Can be a positive number for a shift towards the downstream region
        (towards the inside of the aligned read) and a negative number for a shift towards the upstream region.
        reverse_shift -- Number of bps to shift the reads aligned to the reverse strand.
        Can be a positive number for a shift towards the upstream region and a negative number
        for a shift towards the downstream region (towards the inside of the aligned read).

        Return:
        hon_signal -- Normalized signal.
        slopehon_signal -- Slope signal.
        """

        # Fetch raw signal
        pileup_region = PileupRegion(start, end, downstream_ext, upstream_ext, forward_shift, reverse_shift)
        if (ps_version == "0.7.5"):
            self.bam.fetch(reference=ref, start=start, end=end, callback=pileup_region)
        else:
            iter = self.bam.fetch(reference=ref, start=start, end=end)
            for alignment in iter:
                pileup_region.__call__(alignment)
        raw_signal = array([min(e, initial_clip) for e in pileup_region.vector])

        # Std-based clipping
        mean = raw_signal.mean()
        std = raw_signal.std()
        clip_signal = [min(e, mean + (10 * std)) for e in raw_signal]

        # Cleavage bias correction
        bias_corrected_signal = self.bias_correction(clip_signal, bias_table, genome_file_name,
                                                     ref, start, end, forward_shift, reverse_shift, strands_specific)

        # Boyle normalization (within-dataset normalization)
        boyle_signal = array(self.boyle_norm(bias_corrected_signal))

        # Hon normalization (between-dataset normalization)
        perc = scoreatpercentile(boyle_signal, per_norm)
        std = boyle_signal.std()
        hon_signal = self.hon_norm(boyle_signal, perc, std)

        # Slope signal
        slope_signal = self.slope(hon_signal, self.sg_coefs)

        # Hon normalization on slope signal (between-dataset slope smoothing)
        abs_seq = array([abs(e) for e in slope_signal])
        perc = scoreatpercentile(abs_seq, per_slope)
        std = abs_seq.std()
        slopehon_signal = self.hon_norm(slope_signal, perc, std)

        # Writing signal
        if (print_raw_signal):
            signal_file = open(print_raw_signal, "a")
            signal_file.write("fixedStep chrom=" + ref + " start=" + str(start + 1) + " step=1\n" + "\n".join(
                [str(e) for e in nan_to_num(raw_signal)]) + "\n")
            signal_file.close()
        if (print_bc_signal):
            signal_file = open(print_bc_signal, "a")
            signal_file.write("fixedStep chrom=" + ref + " start=" + str(start + 1) + " step=1\n" + "\n".join(
                [str(e) for e in nan_to_num(bias_corrected_signal)]) + "\n")
            signal_file.close()
        if (print_norm_signal):
            signal_file = open(print_norm_signal, "a")
            signal_file.write("fixedStep chrom=" + ref + " start=" + str(start + 1) + " step=1\n" + "\n".join(
                [str(e) for e in nan_to_num(hon_signal)]) + "\n")
            signal_file.close()
        if (print_slope_signal):
            signal_file = open(print_slope_signal, "a")
            signal_file.write("fixedStep chrom=" + ref + " start=" + str(start + 1) + " step=1\n" + "\n".join(
                [str(e) for e in nan_to_num(slope_signal)]) + "\n")
            signal_file.close()

        # Returning normalized and slope sequences
        return hon_signal, slopehon_signal

    def bias_correction(self, signal, bias_table, genome_file_name, chrName, start, end,
                        forward_shift, reverse_shift, strands_specific):
        """
        Performs bias correction.

        Keyword arguments:
        signal -- Input signal.
        bias_table -- Bias table.

        Return:
        bias_corrected_signal -- Bias-corrected sequence.
        """

        if (not bias_table): return signal

        # Parameters
        window = 50
        defaultKmerValue = 1.0

        # Initialization
        fastaFile = Fastafile(genome_file_name)
        fBiasDict = bias_table[0]
        rBiasDict = bias_table[1]
        k_nb = len(fBiasDict.keys()[0])
        p1 = start
        p2 = end
        p1_w = p1 - (window / 2)
        p2_w = p2 + (window / 2)
        p1_wk = p1_w - int(floor(k_nb / 2.))
        p2_wk = p2_w + int(ceil(k_nb / 2.))
        if (p1 <= 0 or p1_w <= 0 or p1_wk <= 0): return signal

        # Raw counts
        nf = [0.0] * (p2_w - p1_w)
        nr = [0.0] * (p2_w - p1_w)
        for read in self.bam.fetch(chrName, p1_w, p2_w):
            if (not read.is_reverse):
                cut_site = read.pos + forward_shift
                if cut_site >= start and cut_site < end:
                    nf[cut_site - p1_w] += 1.0
                    # for i in range(max(read.pos + forward_shift, start), min(read.pos + forward_shift + 1, end - 1)):
                    #    nf[i - start] += 1.0
            else:
                cut_site = read.aend + reverse_shift - 1
                if cut_site >= start and cut_site < end:
                    nr[cut_site - p1_w] += 1.0
                    # for i in range(max(read.aend + reverse_shift - 1, start), min(read.aend + reverse_shift, end - 1)):
                    #    nr[i - start] += 1.0

                    # if ((not read.is_reverse) and (read.pos > p1_w)): nf[read.pos - p1_w] += 1.0
                    # if ((read.is_reverse) and ((read.aend - 1) < p2_w)): nr[read.aend - 1 - p1_w] += 1.0

        # Smoothed counts
        Nf = []
        Nr = []
        fSum = sum(nf[:window])
        rSum = sum(nr[:window])
        fLast = nf[0]
        rLast = nr[0]
        for i in range((window / 2), len(nf) - (window / 2)):
            Nf.append(fSum)
            Nr.append(rSum)
            fSum -= fLast
            fSum += nf[i + (window / 2)]
            fLast = nf[i - (window / 2) + 1]
            rSum -= rLast
            rSum += nr[i + (window / 2)]
            rLast = nr[i - (window / 2) + 1]

        # Fetching sequence
        currStr = str(fastaFile.fetch(chrName, p1_wk-1, p2_wk-2)).upper()
        currRevComp = AuxiliaryFunctions.revcomp(str(fastaFile.fetch(chrName,p1_wk+2, p2_wk+1)).upper())
        #currStr = str(fastaFile.fetch(chrName, p1_wk, p2_wk - 1)).upper()
        #currRevComp = AuxiliaryFunctions.revcomp(str(fastaFile.fetch(chrName, p1_wk + 1,
         #                                                            p2_wk)).upper())

        # Iterating on sequence to create signal
        af = []
        ar = []
        for i in range(int(ceil(k_nb / 2.)), len(currStr) - int(floor(k_nb / 2)) + 1):
            fseq = currStr[i - int(floor(k_nb / 2.)):i + int(ceil(k_nb / 2.))]
            rseq = currRevComp[len(currStr) - int(ceil(k_nb / 2.)) - i:len(currStr) + int(floor(k_nb / 2.)) - i]
            try:
                af.append(fBiasDict[fseq])
            except Exception:
                af.append(defaultKmerValue)
            try:
                ar.append(rBiasDict[rseq])
            except Exception:
                ar.append(defaultKmerValue)

        # Calculating bias and writing to wig file
        fSum = sum(af[:window])
        rSum = sum(ar[:window])
        fLast = af[0]
        rLast = ar[0]
        bias_corrected_signal = []
        bias_corrected_signal_forward = []
        bias_corrected_signal_reverse = []
        for i in range((window / 2), len(af) - (window / 2)):
            nhatf = Nf[i - (window / 2)] * (af[i] / fSum)
            nhatr = Nr[i - (window / 2)] * (ar[i] / rSum)
            zf = log(nf[i] + 1) - log(nhatf + 1)
            zr = log(nr[i] + 1) - log(nhatr + 1)
            bias_corrected_signal_forward.append(zf)
            bias_corrected_signal_reverse.append(zr)
            bias_corrected_signal.append(zf + zr)
            fSum -= fLast
            fSum += af[i + (window / 2)]
            fLast = af[i - (window / 2) + 1]
            rSum -= rLast
            rSum += ar[i + (window / 2)]
            rLast = ar[i - (window / 2) + 1]

        # Fixing the negative number in bias corrected signal
        min_value = abs(min(bias_corrected_signal_forward))
        bias_fixed_signal_forward = [e + min_value for e in bias_corrected_signal_forward]

        min_value = abs(min(bias_corrected_signal_reverse))
        bias_fixed_signal_reverse = [e + min_value for e in bias_corrected_signal_reverse]

        min_value = abs(min(bias_corrected_signal))
        bias_fixed_signal = [e + min_value for e in bias_corrected_signal]

        # Termination
        fastaFile.close()
        if not strands_specific:
            return bias_corrected_signal
        else:
            return bias_fixed_signal_forward, bias_fixed_signal_reverse

    def hon_norm(self, sequence, mean, std):
        """
        Normalizes a sequence according to hon's criterion using mean and std.
        This represents a between-dataset normalization.

        Keyword arguments:
        sequence -- Input sequence.
        mean -- Global mean.
        std -- Global std.

        Return:
        norm_seq -- Normalized sequence.
        """

        #if std != 0:
        #    norm_seq = []
        #    for e in sequence:
        #        norm_seq.append(1.0 / (1.0 + (exp(-(e - mean) / std))))
        #    return norm_seq
        #else:
        #    return sequence
        norm_seq = []
        for e in sequence:
            if(e == 0.0): norm_seq.append(0.0)
            elif(e > 0.0): norm_seq.append(1.0/(1.0+(exp(-(e-mean)/std))))
            else: norm_seq.append(-1.0/(1.0+(exp(-(-e-mean)/std))))
        return norm_seq

    def boyle_norm(self, sequence):
        """
        Normalizes a sequence according to Boyle's criterion.
        This represents a within-dataset normalization.

        Keyword arguments:
        sequence -- Input sequence.

        Return:
        norm_seq -- Normalized sequence.
        """
        mean = array([e for e in sequence if e > 0]).mean()
        if isnan(mean):
            return sequence
        else:
            norm_seq = [(float(e) / mean) for e in sequence]
            return norm_seq

    def savitzky_golay_coefficients(self, window_size, order, deriv):
        """
        Evaluate the Savitzky-Golay coefficients in order to evaluate the slope of the signal.
        It uses a window_size (of the interpolation), order (of the polynomial), deriv (derivative needed).

        Keyword arguments:
        window_size -- Size of the window for function interpolation.
        order -- Order of polynomial.
        deriv -- Derivative.

        Return:
        m[::-1] -- The Savitzky-Golay coefficients.
        """

        # Get statistics
        # try: # TODO ERRORS
        window_size = abs(int(window_size))
        order = abs(int(order))
        # except ValueError, msg:
        #    raise ValueError("windowSize and order have to be of type int")
        # if windowSize % 2 != 1 or windowSize < 1:
        #    raise TypeError("windowSize size must be a positive odd number")
        # if windowSize < order + 2:
        #    raise TypeError("windowSize is too small for the polynomials order")
        order_range = range(order + 1)
        half_window = (window_size - 1) // 2

        # Precompute Coefficients
        b = mat([[k ** i for i in order_range] for k in range(-half_window, half_window + 1)])
        m = linalg.pinv(b).A[deriv]
        return m[::-1]

    def slope(self, sequence, sg_coefs):
        """
        Evaluates the slope of sequence given the sg_coefs loaded.

        Keyword arguments:
        sequence -- Input sequence.
        sg_coefs -- Savitzky-Golay coefficients.

        Return:
        slope_seq -- Slope sequence.
        """
        slope_seq = convolve(sequence, sg_coefs)
        slope_seq = [e for e in slope_seq[(len(sg_coefs) / 2):(len(slope_seq) - (len(sg_coefs) / 2))]]

        return slope_seq

    def get_signal_per_strand(self, ref, start, end, downstream_ext, upstream_ext, forward_shift, reverse_shift,
                              initial_clip=1000, per_norm=98, per_slope=98,
                              bias_table=None, genome_file_name=None, print_raw_signal=False,
                              print_bc_signal=False, print_norm_signal=False, print_slope_signal=False,
                              strands_specific=True):
        """

        :param ref: Chromosome name.
        :param start: Initial genomic coordinate of signal.
        :param end: Final genomic coordinate of signal.
        :param downstream_ext: Number of bps to extend towards the downstream region
        :param upstream_ext: Number of bps to extend towards the upstream region
        :param forward_shift: Number of bps to shift the reads aligned to the forward strand.
        :param reverse_shift: Number of bps to shift the reads aligned to the reverse strand.
        :param initial_clip: Signal will be initially clipped at this level to avoid outliers.
        :param per_norm: Percentile value for 'hon_norm' function of the normalized signal.
        :param per_slope: Percentile value for 'hon_norm' function of the slope signal.
        :param bias_table: Bias table to perform bias correction.
        :param genome_file_name: Genome to perform bias correction.
        :param print_raw_signal:
        :param print_bc_signal:
        :param print_norm_signal:
        :param print_slope_signal:
        :return: normalized and slope signal for each strand.
        """

        raw_signal_forward = [0.0] * (end - start)
        raw_signal_reverse = [0.0] * (end - start)

        reads = self.bam.fetch(reference=ref, start=start, end=end)
        for read in reads:
            if (not read.is_reverse):
                cut_site = read.pos + forward_shift
                if cut_site >= start and cut_site < end:
                    raw_signal_forward[cut_site - start] += 1.0
            else:
                cut_site = read.aend + reverse_shift - 1
                if cut_site >= start and cut_site < end:
                    raw_signal_reverse[cut_site - start] += 1.0

        raw_signal_forward = array([min(e, initial_clip) for e in raw_signal_forward])
        raw_signal_reverse = array([min(e, initial_clip) for e in raw_signal_reverse])

        # Std-based clipping
        mean = raw_signal_forward.mean()
        std = raw_signal_forward.std()
        clip_signal_forward = [min(e, mean + (10 * std)) for e in raw_signal_forward]
        mean = raw_signal_reverse.mean()
        std = raw_signal_reverse.std()
        clip_signal_reverse = [min(e, mean + (10 * std)) for e in raw_signal_reverse]

        # Cleavage bias correction
        bc_signal_forward = None
        bc_signal_reverse = None
        if bias_table:
            bc_signal_forward, bc_signal_reverse = self.bias_correction(raw_signal_forward, bias_table,
                                                                        genome_file_name,
                                                                        ref, start, end, forward_shift, reverse_shift,
                                                                        strands_specific)
        else:
            bc_signal_forward = clip_signal_forward
            bc_signal_reverse = clip_signal_reverse

        # Boyle normalization (within-dataset normalization)
        boyle_signal_forward = array(self.boyle_norm(bc_signal_forward))
        boyle_signal_reverse = array(self.boyle_norm(bc_signal_reverse))

        # Hon normalization (between-dataset normalization)
        perc = scoreatpercentile(boyle_signal_forward, per_norm)
        std = boyle_signal_forward.std()
        hon_signal_forward = self.hon_norm(boyle_signal_forward, perc, std)

        perc = scoreatpercentile(boyle_signal_reverse, per_norm)
        std = boyle_signal_reverse.std()
        hon_signal_reverse = self.hon_norm(boyle_signal_reverse, perc, std)

        # Slope signal
        slope_signal_forward = self.slope(hon_signal_forward, self.sg_coefs)
        slope_signal_reverse = self.slope(hon_signal_reverse, self.sg_coefs)

        # Returning normalized and slope sequences
        return hon_signal_forward, slope_signal_forward, hon_signal_reverse, slope_signal_reverse


from __future__ import print_function
import os
from rgt.GenomicRegion import GenomicRegion
from rgt.GenomicRegionSet import GenomicRegionSet

####################################################################################
####################################################################################

class BindingSite(GenomicRegion):
    """Describes a binding region on DNA or RNA including the information regarding to this region."""
    __slots__ = ['name', 'score', 'errors_bp', 'motif', 'orientation', 'seq' ]

    def __init__(self, chrom, initial, final, name=None, score=None, errors_bp=None, motif=None, 
                 strand=None, orientation=None, guanine_rate=None, seq=None):
        """*Keyword arguments:*
        
            - name -- The name of this binding site (Default: None)
            - seq_type -- DNA or RNA
            - chrm -- Define the chromosome for DNA; for RNA its default is "RNA"
            - initial -- Binding start position
            - final -- Binding end position
            - score -- Score of the binding pattern (Default: None)
            - errors_bp -- Error base pair in this binding (Default: None)
            - motif -- The motif for this binding (Default: None)
            - strand -- The strand of DNA (+ or -) (Default: None)
            - orientation -- Parallel or antiparallel (Default: None)
            - guanine_rate -- (Default: None)
            - seq -- Sequence of this region with ATCG as letters
        """
        GenomicRegion.__init__(self, chrom=chrom, initial=initial, final=final)
        
        self.name = name                      # RNA name
        self.score = score                    # Score for pattern matching
        self.errors_bp = errors_bp                  
        self.motif = motif
        #self.strand = strand
        self.orientation = orientation
        self.seq = seq                        # An object (Sequence) not just a string
        if seq:
            self.guanine_rate = "{0:.2f}".format(float(seq.seq.count("G"))/len(seq))

    def __str__(self):
        """Give informal string representation."""
        infos = [ self.name, self.chrom, self.initial, self.final, self.score, self.errors_bp, 
                  self.motif, self.orientation, self.seq ]
        return '-'.join( [str(x) for x in infos if x] )
        
    def __repr__(self):
        """Return official representation of GenomicRegion."""
        infos = [ self.name, self.chrom, self.initial, self.final, self.score, self.errors_bp, 
                  self.motif, self.orientation, self.seq ]
        return ','.join( [str(x) for x in infos if x] ) 
        
    def __len__(self):
        """Return the length of the binding site."""
        return self.final - self.initial

    def __eq__(self, other):
        return (self.initial, self.final) == (other.initial, other.final)

    def __hash__(self):
        return hash(tuple([self.chrom, self.initial, self.final]))
                        
    def str_rna(self, pa=True):
        if pa:
            return "{0}-{1}-{2}".format(self.initial, self.final, self.orientation)
        else:
            return "{0}-{1}".format(self.initial, self.final)
            
            
            
            

####################################################################################
####################################################################################

class BindingSiteSet(GenomicRegionSet):
    """Represent a collection of RNABinding with some functions."""

    def __init__(self, name):
        """*Keyword arguments:*

            - name -- Define the name of this BindingSiteSet.
        """
        GenomicRegionSet.__init__(self, name = name)

    def sort(self):
        """Sort Elements by criteria defined by a GenomicRegion."""
        self.sequences.sort(cmp = GenomicRegion.__cmp__)
        self.sorted = True
    
    def get_bs(self, orientation):
        """Get the Binding Sites with the given orientation.

        *Keyword arguments:*

            - orientation -- Define the orientation of the sequence.
        """
        output = BindingSiteSet(self.name+":"+orientation)
        for bs in self.sequences:
            if bs.orientation == orientation:
                output.add(bs)
        return output

    def write_rbs(self, filename):
        """Write the information into a file with .rbs """
        d = os.path.dirname(filename)
        if not os.path.exists(d):
            os.makedirs(d)

        with open(filename, "w") as f:
            print("# Sequence-ID\tStart\tEnd\tScore\tMotif\tError-rate\tGuanine-rate\tSequence",file=f)
            # Sequence-ID   Start   End Score   Motif   Error-rate  Errors  
            # Guanine-rate    Duplicates  TFO Duplicate locations

            for bs in self.sequences:
                err_rate = 1 - bs.score/(bs.final - bs.initial)
                print("\t".join([bs.name, str(bs.initial), str(bs.final), str(bs.score), 
                                 bs.motif, "{0:.2f}".format(err_rate), bs.guanine_rate, bs.seq.seq]), file=f)

    def write_dbs(self, filename):
        """Write the information into a file with .dbs """
        d = os.path.dirname(filename)
        if not os.path.exists(d):
            os.makedirs(d)

        with open(filename, "w") as f:
            print("# Chromosome\tStart\tEnd\tScore\tError-rate\tGuanine-rate\tSequence",file=f)
            # Sequence-ID   Start   End Score   Motif   Error-rate  Errors  
            # Guanine-rate    Duplicates  TFO Duplicate locations

            for bs in self.sequences:
                print(bs)
                err_rate = 1 - bs.score/(bs.final - bs.initial)
                print("\t".join([bs.chrom, str(bs.initial), str(bs.final), str(bs.score), 
                                 "{0:.2f}".format(err_rate), bs.guanine_rate, bs.seq.seq]), file=f)

    def concatenate(self, another_BindingSiteSet):
        """Concatenate another RNABindingSet without sorting"""
        self.sequences = self.sequences + another_BindingSiteSet.sequences

    def count_rbs_position(self, bp):
        count = 0
        for r in self.sequences:
            if r.initial <= bp and r.final > bp:
                count += 1
        return count

